#!/usr/bin/env python
"""
XANES Data Viewer and Analysis Tool
"""
import os
import sys
import time
import copy
import platform
import numpy as np
np.seterr(all='ignore')

from functools import partial
from collections import OrderedDict
import wx
import wx.lib.scrolledpanel as scrolled

import wx.lib.mixins.inspection

from wx.richtext import RichTextCtrl

is_wxPhoenix = 'phoenix' in wx.PlatformInfo
is_windows = platform.system().startswith('Windows')
WX_DEBUG = True

import larch
from larch import Group
from larch.math import index_of
from larch.utils.strutils import (file2groupname, unique_name,
                                  common_startstring)

from larch.larchlib import read_workdir, save_workdir, read_config, save_config

from larch.wxlib import (LarchFrame, ColumnDataFileFrame, AthenaImporter,
                         FileCheckList, FloatCtrl, SetTip, get_icon,
                         SimpleText, pack, Button, Popup, HLine, FileSave,
                         Choice, Check, MenuItem, GUIColors, CEN,
                         LEFT, FRAMESTYLE, Font, FONTSIZE, flatnotebook)

from larch.wxlib.plotter import _newplot, _plot, last_cursor_pos

from larch.fitting import fit_report
from larch.utils import group2dict
from larch.site_config import icondir

from .prepeak_panel import PrePeakPanel
from .xasnorm_panel import XASNormPanel
from .lincombo_panel import LinearComboPanel
from .pca_panel import PCAPanel
from .exafs_panel import EXAFSPanel
from .regress_panel import RegressionPanel

from .xas_dialogs import (MergeDialog, RenameDialog, RemoveDialog,
                          DeglitchDialog, ExportCSVDialog, RebinDataDialog,
                          EnergyCalibrateDialog, SmoothDataDialog,
                          OverAbsorptionDialog, DeconvolutionDialog,
                          SpectraCalcDialog,  QuitDialog)

from larch.io import (read_ascii, read_xdi, read_gsexdi,
                      gsescan_group, fix_varname, groups2csv,
                      is_athena_project, AthenaProject, make_hashkey)

from larch.xafs import pre_edge, pre_edge_baseline

LEFT = wx.ALIGN_LEFT
CEN |=  wx.ALL
FILE_WILDCARDS = "Data Files(*.0*,*.dat,*.xdi,*.prj)|*.0*;*.dat;*.xdi;*.prj|All files (*.*)|*.*"

ICON_FILE = 'onecone.ico'
XASVIEW_SIZE = (950, 750)
PLOTWIN_SIZE = (550, 550)

NB_PANELS = OrderedDict((('Normalization', XASNormPanel),
                         ('Pre-edge Peak', PrePeakPanel),
                         ('PCA',  PCAPanel),
                         ('Linear Combo', LinearComboPanel),
                         ('Regression', RegressionPanel),
                         ('EXAFS', EXAFSPanel)))

QUIT_MESSAGE = '''Really Quit? You may want to save your project before quitting.
 This is not done automatically!'''


def assign_gsescan_groups(group):
    labels = group.array_labels
    labels = []
    for i, name in enumerate(group.pos_desc):
        name = fix_varname(name.lower())
        labels.append(name)
        setattr(group, name, group.pos[i, :])

    for i, name in enumerate(group.sums_names):
        name = fix_varname(name.lower())
        labels.append(name)
        setattr(group, name, group.sums_corr[i, :])

    for i, name in enumerate(group.det_desc):
        name = fix_varname(name.lower())
        labels.append(name)
        setattr(group, name, group.det_corr[i, :])

    group.array_labels = labels

class XASController():
    """
    class hollding the Larch session and doing the
    processing work for Larch XAS GUI
    """
    config_file = 'xas_viewer.conf'
    def __init__(self, wxparent=None, _larch=None):
        self.wxparent = wxparent
        self.larch = _larch
        if self.larch is None:
            self.larch = larch.Interpreter()
        self.filelist = None
        self.file_groups = OrderedDict()
        self.fit_opts = {}
        self.group = None

        self.groupname = None
        self.report_frame = None
        self.symtable = self.larch.symtable

    def init_larch(self):
        _larch = self.larch
        old_config = read_config(self.config_file)

        config = self.make_default_config()
        for sname in config:
            if old_config is not None and sname in old_config:
                val = old_config[sname]
                if isinstance(val, dict):
                    config[sname].update(val)
                else:
                    config[sname] = val

        for key, value in config.items():
            setattr(_larch.symtable._sys.xas_viewer, key, value)
        try:
            os.chdir(config['workdir'])
        except:
            pass
        self.wxparent.SetIcon(wx.Icon(self.get_iconfile(), wx.BITMAP_TYPE_ICO))

    def make_default_config(self):
        """ default config, probably called on first run of program"""
        config = {'chdir_on_fileopen': True,
                  'workdir': os.getcwd()}
        return config

    def get_config(self, key, default=None):
        "get configuration setting"
        confgroup = self.larch.symtable._sys.xas_viewer
        return getattr(confgroup, key, default)

    def save_config(self):
        """save configuration"""
        conf = group2dict(self.larch.symtable._sys.xas_viewer)
        conf.pop('__name__')
        save_config(self.config_file, conf)

    def set_workdir(self):
        self.larch.symtable._sys.xas_viewer.workdir = os.getcwd()

    def get_iconfile(self):
        return os.path.join(icondir, ICON_FILE)

    def write_message(self, msg, panel=0):
        """write a message to the Status Bar"""
        self.wxparent.statusbar.SetStatusText(msg, panel)

    def close_all_displays(self):
        "close all displays, as at exit"
        self.symtable._plotter.close_all_displays()

    def get_display(self, win=1, stacked=False):
        wintitle='Larch XAS Plot Window %i' % win
        # if stacked:
        #     win = 2
        #    wintitle='Larch XAS Plot Window'
        opts = dict(wintitle=wintitle, stacked=stacked, win=win,
                    size=PLOTWIN_SIZE)
        out = self.symtable._plotter.get_display(**opts)
        if win > 1:
            p1 = getattr(self.symtable._plotter, 'plot1', None)
            if p1 is not None:
                try:
                    siz = p1.GetSize()
                    pos = p1.GetPosition()
                    pos[0] += int(siz[0]/4)
                    pos[1] += int(siz[1]/4)
                    out.SetSize(pos)
                    if not stacked:
                        out.SetSize(siz)
                except Exception:
                    pass
        return out

    def get_group(self, groupname=None):
        if groupname is None:
            groupname = self.groupname
            if groupname is None:
                return None
        dgroup = getattr(self.symtable, groupname, None)
        if dgroup is None and groupname in self.file_groups:
            groupname = self.file_groups[groupname]
            dgroup = getattr(self.symtable, groupname, None)
        return dgroup

    def filename2group(self, filename):
        "convert filename (as displayed) to larch group"
        return self.get_group(self.file_groups[str(filename)])

    def merge_groups(self, grouplist, master=None, yarray='mu', outgroup=None):
        """merge groups"""
        cmd = """%s = merge_groups(%s, master=%s,
        xarray='energy', yarray='%s', kind='cubic', trim=True)
        """
        glist = "[%s]" % (', '.join(grouplist))
        outgroup = fix_varname(outgroup.lower())
        if outgroup is None:
            outgroup = 'merged'

        outgroup = unique_name(outgroup, self.file_groups, max=1000)

        cmd = cmd % (outgroup, glist, master, yarray)
        self.larch.eval(cmd)

        if master is None:
            master = grouplist[0]
        this = self.get_group(outgroup)
        master = self.get_group(master)
        if not hasattr(this, 'xasnorm_config'):
            this.xasnorm_config = {}
        this.xasnorm_config.update(master.xasnorm_config)
        this.datatype = master.datatype
        this.xdat = 1.0*this.energy
        this.ydat = 1.0*getattr(this, yarray)
        this.yerr =  getattr(this, 'd' + yarray, 1.0)
        if yarray != 'mu':
            this.mu = this.ydat
        this.plot_xlabel = 'energy'
        this.plot_ylabel = yarray
        return outgroup

    def copy_group(self, filename, new_filename=None):
        """copy XAS group (by filename) to new group"""
        groupname = self.file_groups[filename]
        if not hasattr(self.larch.symtable, groupname):
            return

        ogroup = self.get_group(groupname)

        ngroup = Group(datatype=ogroup.datatype,
                       copied_from=groupname)

        for attr in dir(ogroup):
            do_copy = True
            if attr in ('xdat', 'ydat', 'i0', 'data' 'yerr',
                        'energy', 'mu'):
                val = getattr(ogroup, attr)*1.0
            elif attr in ('norm', 'flat', 'deriv', 'deconv',
                          'post_edge', 'pre_edge', 'norm_mback',
                          'norm_vict', 'norm_poly'):
                do_copy = False
            else:
                try:
                    val = copy.deepcopy(getattr(ogroup, attr))
                except ValueError:
                    do_copy = False
            if do_copy:
                setattr(ngroup, attr, val)

        if new_filename is None:
            new_filename = filename + '_1'
        ngroup.filename = unique_name(new_filename, self.file_groups.keys())
        ngroup.groupname = unique_name(groupname, self.file_groups.values())
        setattr(self.larch.symtable, ngroup.groupname, ngroup)
        return ngroup

    def get_cursor(self, win=None):
        """get last cursor from selected window"""
        return last_cursor_pos(win=win, _larch=self.larch)

    def plot_group(self, groupname=None, title=None, plot_yarrays=None,
                   new=True, zoom_out=True, **kws):
        ppanel = self.get_display(stacked=False).panel
        newplot = ppanel.plot
        oplot   = ppanel.oplot
        plotcmd = oplot
        viewlims = ppanel.get_viewlimits()
        if new:
            plotcmd = newplot

        dgroup = self.get_group(groupname)
        if not hasattr(dgroup, 'xdat'):
            print("Cannot plot group ", groupname)

        if ((getattr(dgroup, 'plot_yarrays', None) is None or
             getattr(dgroup, 'energy', None) is None or
             getattr(dgroup, 'mu', None) is None)):
            self.process(dgroup)

        if plot_yarrays is None and hasattr(dgroup, 'plot_yarrays'):
            plot_yarrays = dgroup.plot_yarrays

        popts = kws
        path, fname = os.path.split(dgroup.filename)
        if not 'label' in popts:
            popts['label'] = dgroup.plot_ylabel
        zoom_out = (zoom_out or
                  min(dgroup.xdat) >= viewlims[1] or
                  max(dgroup.xdat) <= viewlims[0] or
                  min(dgroup.ydat) >= viewlims[3] or
                  max(dgroup.ydat) <= viewlims[2])

        if not zoom_out:
            popts['xmin'] = viewlims[0]
            popts['xmax'] = viewlims[1]
            popts['ymin'] = viewlims[2]
            popts['ymax'] = viewlims[3]

        popts['xlabel'] = dgroup.plot_xlabel
        popts['ylabel'] = dgroup.plot_ylabel
        if getattr(dgroup, 'plot_y2label', None) is not None:
            popts['y2label'] = dgroup.plot_y2label

        plot_extras = None
        if new:
            if title is None:
                title = fname
            plot_extras = getattr(dgroup, 'plot_extras', None)

        popts['title'] = title
        if hasattr(dgroup, 'custom_plotopts'):
            popts.update(dgroup.custom_plotopts)

        narr = len(plot_yarrays) - 1
        for i, pydat in enumerate(plot_yarrays):
            yaname, yopts, yalabel = pydat
            popts.update(yopts)
            if yalabel is not None:
                popts['label'] = yalabel
            popts['delay_draw'] = (i != narr)

            plotcmd(dgroup.xdat, getattr(dgroup, yaname), **popts)
            plotcmd = oplot

        if plot_extras is not None:
            axes = ppanel.axes
            for etype, x, y, opts in plot_extras:
                if etype == 'marker':
                    popts = {'marker': 'o', 'markersize': 4,
                             'label': '_nolegend_',
                             'markerfacecolor': 'red',
                             'markeredgecolor': '#884444'}
                    popts.update(opts)
                    axes.plot([x], [y], **popts)
                elif etype == 'vline':
                    popts = {'ymin': 0, 'ymax': 1.0,
                             'color': '#888888'}
                    popts.update(opts)
                    axes.axvline(x, **popts)
        ppanel.canvas.draw()


class XASFrame(wx.Frame):
    _about = """Larch XAS GUI: XAS Visualization and Analysis

    Matt Newville <newville @ cars.uchicago.edu>
    """
    def __init__(self, parent=None, _larch=None, filename=None, **kws):
        wx.Frame.__init__(self, parent, -1, size=XASVIEW_SIZE, style=FRAMESTYLE)

        self.last_array_sel = {}
        self.paths2read = []

        title = "Larch XAS GUI: XAS Visualization and Analysis"

        self.larch_buffer = parent
        if not isinstance(parent, LarchFrame):
            self.larch_buffer = LarchFrame(_larch=_larch, is_standalone=False)

        self.larch_buffer.Show()
        self.larch_buffer.Raise()
        self.larch = self.larch_buffer.larchshell
        self.larch.symtable._sys.xas_viewer = Group()

        self.controller = XASController(wxparent=self, _larch=self.larch)
        self.subframes = {}
        self.plotframe = None
        self.SetTitle(title)
        self.SetSize(XASVIEW_SIZE)

        self.SetFont(Font(FONTSIZE))
        self.larch_buffer.Hide()
        self.createMainPanel()
        self.createMenus()
        self.statusbar = self.CreateStatusBar(2, style=wx.STB_DEFAULT_STYLE)
        self.statusbar.SetStatusWidths([-3, -1])
        statusbar_fields = [" ", "initializing...."]
        for i in range(len(statusbar_fields)):
            self.statusbar.SetStatusText(statusbar_fields[i], i)
            self.current_filename = filename
        self.Show()
        if filename is not None:
            wx.CallAfter(self.onRead, filename)

    def createMainPanel(self):
        display0 = wx.Display(0)
        client_area = display0.ClientArea
        xmin, ymin, xmax, ymax = client_area
        xpos = int((xmax-xmin)*0.02) + xmin
        ypos = int((ymax-ymin)*0.04) + ymin
        self.SetPosition((xpos, ypos))

        splitter  = wx.SplitterWindow(self, style=wx.SP_LIVE_UPDATE)
        splitter.SetMinimumPaneSize(250)

        leftpanel = wx.Panel(splitter)
        ltop = wx.Panel(leftpanel)

        def Btn(msg, x, act):
            b = Button(ltop, msg, size=(x, 30),  action=act)
            b.SetFont(Font(FONTSIZE))
            return b

        sel_none = Btn('Select None',   120, self.onSelNone)
        sel_all  = Btn('Select All',    120, self.onSelAll)

        self.controller.filelist = FileCheckList(leftpanel,
                                                 select_action=self.ShowFile,
                                                 remove_action=self.RemoveFile)
        tsizer = wx.BoxSizer(wx.HORIZONTAL)
        tsizer.Add(sel_all, 1, LEFT|wx.GROW, 1)
        tsizer.Add(sel_none, 1, LEFT|wx.GROW, 1)
        pack(ltop, tsizer)

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(ltop, 0, LEFT|wx.GROW, 1)
        sizer.Add(self.controller.filelist, 1, LEFT|wx.GROW|wx.ALL, 1)

        pack(leftpanel, sizer)

        # right hand side
        panel = wx.Panel(splitter)
        sizer = wx.BoxSizer(wx.VERTICAL)

        self.title = SimpleText(panel, 'initializing...', size=(300, -1))
        self.title.SetFont(Font(FONTSIZE+2))

        ir = 0
        sizer.Add(self.title, 0, LEFT|wx.GROW|wx.ALL, 1)
        self.nb = flatnotebook(panel, NB_PANELS,
                               panelkws=dict(controller=self.controller),
                               on_change=self.onNBChanged)
        sizer.Add(self.nb, 1, LEFT|wx.EXPAND, 2)
        pack(panel, sizer)

        splitter.SplitVertically(leftpanel, panel, 1)
        wx.CallAfter(self.init_larch)

    def onNBChanged(self, event=None):
        callback = getattr(self.nb.GetCurrentPage(), 'onPanelExposed', None)
        if callable(callback):
            callback()

    def onSelAll(self, event=None):
        self.controller.filelist.select_all()

    def onSelNone(self, event=None):
        self.controller.filelist.select_none()

    def init_larch(self):
        self.SetStatusText('initializing Larch', 1)
        self.title.SetLabel('')

        self.controller.init_larch()

        plotframe = self.controller.get_display(stacked=False)
        xpos, ypos = self.GetPosition()
        xsiz, ysiz = self.GetSize()
        plotframe.SetPosition((xpos+xsiz+5, ypos))
        plotframe.SetSize((600, 650))

        self.SetStatusText('ready', 1)
        self.Raise()

    def write_message(self, msg, panel=0):
        """write a message to the Status Bar"""
        self.statusbar.SetStatusText(msg, panel)

    def RemoveFile(self, fname=None, **kws):
        if fname is not None:
            s = str(fname)
            if s in self.controller.file_groups:
                group = self.controller.file_groups.pop(s)

    def ShowFile(self, evt=None, groupname=None, process=True,
                 plot=True, **kws):
        filename = None
        if evt is not None:
            filename = str(evt.GetString())

        if groupname is None and filename is not None:
            groupname = self.controller.file_groups[filename]

        if not hasattr(self.larch.symtable, groupname):
            return

        dgroup = self.controller.get_group(groupname)
        if dgroup is None:
            return

        if filename is None:
            filename = dgroup.filename
        self.title.SetLabel(filename)
        self.current_filename = filename

        self.controller.group = dgroup
        self.controller.groupname = groupname
        cur_panel = self.nb.GetCurrentPage()
        if process:
            cur_panel.fill_form(dgroup)
            cur_panel.skip_process = True
            cur_panel.process(dgroup=dgroup)
            if plot and hasattr(cur_panel, 'plot'):
                cur_panel.plot(dgroup=dgroup)
            cur_panel.skip_process = False


    def createMenus(self):
        # ppnl = self.plotpanel
        self.menubar = wx.MenuBar()
        fmenu = wx.Menu()
        group_menu = wx.Menu()
        data_menu = wx.Menu()
        # ppeak_menu = wx.Menu()
        m = {}

        MenuItem(self, fmenu, "&Open Data File\tCtrl+O",
                 "Open Data File",  self.onReadDialog)

        MenuItem(self, fmenu, "&Save Project\tCtrl+S",
                 "Save Session to Project File",  self.onSaveProject)

        MenuItem(self, fmenu, "Export Selected Groups to Athena Project",
                 "Export Selected Groups to Athena Project",
                 self.onExportAthena)

        MenuItem(self, fmenu, "Export Selected Groups to CSV",
                 "Export Selected Groups to CSV",  self.onExportCSV)

        fmenu.AppendSeparator()

        MenuItem(self, fmenu, 'Show Larch Buffer\tCtrl+L',
                 'Show Larch Programming Buffer',
                 self.onShowLarchBuffer)

        MenuItem(self, fmenu, 'Save Larch Script of History\tCtrl+H',
                 'Save Session History as Larch Script',
                 self.onSaveLarchHistory)

        if WX_DEBUG:
            MenuItem(self, fmenu, "&Inspect \tCtrl+J",
                     " wx inspection tool ",  self.showInspectionTool)

        MenuItem(self, fmenu, "&Quit\tCtrl+Q", "Quit program", self.onClose)


        MenuItem(self, group_menu, "Copy This Group",
                 "Copy This Group", self.onCopyGroup)

        MenuItem(self, group_menu, "Rename This Group",
                 "Rename This Group", self.onRenameGroup)

        MenuItem(self, group_menu, "Remove Selected Groups",
                 "Remove Selected Group", self.onRemoveGroups)


        group_menu.AppendSeparator()

        MenuItem(self, group_menu, "Merge Selected Groups",
                 "Merge Selected Groups", self.onMergeData)

        group_menu.AppendSeparator()

        MenuItem(self, group_menu, "Freeze Selected Groups",
                 "Freeze Selected Groups", self.onFreezeGroups)

        MenuItem(self, group_menu, "UnFreeze Selected Groups",
                 "UnFreeze Selected Groups", self.onUnFreezeGroups)

        MenuItem(self, data_menu, "Deglitch Data",  "Deglitch Data",
                 self.onDeglitchData)

        MenuItem(self, data_menu, "Recalibrate Energy",
                 "Recalibrate Energy",
                 self.onEnergyCalibrateData)

        MenuItem(self, data_menu, "Smooth Data", "Smooth Data",
                 self.onSmoothData)

        MenuItem(self, data_menu, "Rebin Data", "Rebin Data",
                 self.onRebinData)

        MenuItem(self, data_menu, "Deconvolve Data",
                 "Deconvolution of Data",  self.onDeconvolveData)

        MenuItem(self, data_menu, "Correct Over-absorption",
                 "Correct Over-absorption",
                 self.onCorrectOverAbsorptionData)

        MenuItem(self, data_menu, "Add and Subtract Sepctra",
                 "Calculations of Spectra",  self.onSpectraCalc)

        self.menubar.Append(fmenu, "&File")
        self.menubar.Append(group_menu, "Groups")
        self.menubar.Append(data_menu, "Data")

        # self.menubar.Append(ppeak_menu, "PreEdgePeaks")
        self.SetMenuBar(self.menubar)
        self.Bind(wx.EVT_CLOSE,  self.onClose)

    def onShowLarchBuffer(self, evt=None):
        if self.larch_buffer is None:
            self.larch_buffer = LarchFrame(_larch=self.larch, is_standalone=False)
        self.larch_buffer.Show()
        self.larch_buffer.Raise()

    def onSaveLarchHistory(self, evt=None):
        wildcard = 'Larch file (*.lar)|*.lar|All files (*.*)|*.*'
        path = FileSave(self, message='Save Session History as Larch Script',
                        wildcard=wildcard,
                        default_file='xas_viewer_history.lar')
        if path is not None:
            self.larch._larch.input.history.save(path, session_only=True)
            self.write_message("Wrote history %s" % path, 0)

    def onExportCSV(self, evt=None):
        filenames = self.controller.filelist.GetCheckedStrings()
        if len(filenames) < 1:
             Popup(self, "No files selected to export to CSV",
                   "No files selected")
             return

        dlg = ExportCSVDialog(self, filenames)
        res = dlg.GetResponse()
        dlg.Destroy()
        if res.ok:
            savegroups = [self.controller.filename2group(res.master)]
            for fname in filenames:
                dgroup = self.controller.filename2group(fname)
                if dgroup not in savegroups:
                    savegroups.append(dgroup)
            groups2csv(savegroups, res.filename, x='energy', y=res.yarray,
                       _larch=self.larch)
            self.write_message("Exported CSV file %s" % (res.filename))

    def onExportAthena(self, evt=None):
        groups = []
        for checked in self.controller.filelist.GetCheckedStrings():
            groups.append(self.controller.file_groups[str(checked)])

        if len(groups) < 1:
             Popup(self, "No files selected to export to Athena",
                   "No files selected")
             return
        self.save_athena_project(groups[0], groups, prompt=True)

    def onSaveProject(self, evt=None):
        groups = [gname for gname in self.controller.file_groups]
        if len(groups) < 1:
            Popup(self, "No files selected to export to Athena",
                  "No files selected")
            return
        self.save_athena_project(groups[0], groups, prompt=True)

    def save_athena_project(self, filename, grouplist, prompt=True):
        if len(grouplist) < 1:
            return
        savegroups = [self.controller.get_group(gname) for gname in grouplist]

        deffile = "%s_%i.prj" % (filename, len(grouplist))
        wcards  = 'Athena Projects (*.prj)|*.prj|All files (*.*)|*.*'

        outfile = FileSave(self, 'Save Groups to Athena Project File',
                           default_file=deffile, wildcard=wcards)

        if outfile is None:
            return
        aprj = AthenaProject(filename=outfile, _larch=self.larch)
        for label, grp in zip(grouplist, savegroups):
            aprj.add_group(grp)

        aprj.save(use_gzip=True)
        self.write_message("Saved project file %s" % (outfile))

    def onConfigDataProcessing(self, event=None):
        pass

    def onNewGroup(self, datagroup):
        """
        install and display a new group, as from 'copy / modify'
        Note: this is a group object, not the groupname or filename
        """
        dgroup = datagroup
        self.install_group(dgroup.groupname, dgroup.filename, overwrite=False)
        self.ShowFile(groupname=dgroup.groupname)

    def onCopyGroup(self, event=None):
        fname = self.current_filename
        if fname is None:
            fname = self.controller.filelist.GetStringSelection()
        ngroup = self.controller.copy_group(fname)
        self.onNewGroup(ngroup)

    def onRenameGroup(self, event=None):
        fname = self.current_filename = self.controller.filelist.GetStringSelection()
        if fname is None:
            return
        dlg = RenameDialog(self, fname)
        res = dlg.GetResponse()
        dlg.Destroy()

        if res.ok:
            selected = []
            for checked in self.controller.filelist.GetCheckedStrings():
                selected.append(str(checked))
            if self.current_filename in selected:
                selected.remove(self.current_filename)
                selected.append(res.newname)

            groupname = self.controller.file_groups.pop(fname)
            self.controller.file_groups[res.newname] = groupname
            self.controller.filelist.rename_item(self.current_filename, res.newname)
            dgroup = self.controller.get_group(groupname)
            dgroup.filename = self.current_filename = res.newname

            self.controller.filelist.SetCheckedStrings(selected)
            self.controller.filelist.SetStringSelection(res.newname)

    def onRemoveGroups(self, event=None):
        groups = []
        for checked in self.controller.filelist.GetCheckedStrings():
            groups.append(str(checked))
        if len(groups) < 1:
            return

        dlg = RemoveDialog(self, groups)
        res = dlg.GetResponse()
        dlg.Destroy()

        if res.ok:
            filelist = self.controller.filelist
            all_fnames = filelist.GetItems()
            for fname in groups:
                gname = self.controller.file_groups.pop(fname)
                delattr(self.controller.symtable, gname)
                all_fnames.remove(fname)

            filelist.Clear()
            for name in all_fnames:
                filelist.Append(name)

    def onFreezeGroups(self, event=None):
        self._freeze_handler(True)

    def onUnFreezeGroups(self, event=None):
        self._freeze_handler(False)

    def _freeze_handler(self, freeze):
        current_filename = self.current_filename
        reproc_group = None
        for fname in self.controller.filelist.GetCheckedStrings():
            groupname = self.controller.file_groups[str(fname)]
            dgroup = self.controller.get_group(groupname)
            if fname == current_filename:
                reproc_group = groupname
            dgroup.is_frozen = freeze
        if reproc_group is not None:
            self.ShowFile(groupname=reproc_group, process=True)

    def onMergeData(self, event=None):
        groups = OrderedDict()
        for checked in self.controller.filelist.GetCheckedStrings():
            cname = str(checked)
            groups[cname] = self.controller.file_groups[cname]
        if len(groups) < 1:
            return

        outgroup = common_startstring(list(groups.keys()))
        outgroup = "%s (merge %d)" % (outgroup, len(groups))
        outgroup = unique_name(outgroup, self.controller.file_groups)
        dlg = MergeDialog(self, list(groups.keys()), outgroup=outgroup)
        res = dlg.GetResponse()
        dlg.Destroy()
        if res.ok:
            fname = res.group
            gname = fix_varname(res.group.lower())
            master = self.controller.file_groups[res.master]
            yname = 'norm' if res.ynorm else 'mu'
            self.controller.merge_groups(list(groups.values()),
                                         master=master,
                                         yarray=yname,
                                         outgroup=gname)
            self.install_group(gname, fname, overwrite=False)
            self.controller.filelist.SetStringSelection(fname)

    def onDeglitchData(self, event=None):
        DeglitchDialog(self, self.controller).Show()

    def onSmoothData(self, event=None):
        SmoothDataDialog(self, self.controller).Show()

    def onRebinData(self, event=None):
        RebinDataDialog(self, self.controller).Show()

    def onCorrectOverAbsorptionData(self, event=None):
        OverAbsorptionDialog(self, self.controller).Show()

    def onSpectraCalc(self, event=None):
        SpectraCalcDialog(self, self.controller).Show()

    def onEnergyCalibrateData(self, event=None):
        EnergyCalibrateDialog(self, self.controller).Show()

    def onDeconvolveData(self, event=None):
        DeconvolutionDialog(self, self.controller).Show()

    def onConfigDataFitting(self, event=None):
        pass

    def showInspectionTool(self, event=None):
        app = wx.GetApp()
        app.ShowInspectionTool()

    def onAbout(self,evt):
        dlg = wx.MessageDialog(self, self._about,
                               "About Larch XAS GUI",
                               wx.OK | wx.ICON_INFORMATION)
        dlg.ShowModal()
        dlg.Destroy()

    def onClose(self, event):
        dlg = QuitDialog(self)
        dlg.Raise()
        dlg.SetWindowStyle(wx.STAY_ON_TOP)
        res = dlg.GetResponse()
        dlg.Destroy()
        if not res.ok:
            return

        if res.save:
            groups = [gname for gname in self.controller.file_groups]
            if len(groups) > 0:
                self.save_athena_project(groups[0], groups, prompt=True)

        self.controller.save_config()
        wx.CallAfter(self.controller.close_all_displays)

        if self.larch_buffer is not None:
            wx.CallAfter(self.larch_buffer.Destroy)

        for name, wid in self.subframes.items():
            if hasattr(wid, 'Destroy'):
                wx.CallAfter(wid.Destroy)

        self.Destroy()

    def show_subframe(self, name, frameclass, **opts):
        shown = False
        if name in self.subframes:
            try:
                self.subframes[name].Raise()
                shown = True
            except:
                del self.subframes[name]
        if not shown:
            self.subframes[name] = frameclass(self, **opts)

    def onSelectColumns(self, event=None):
        dgroup = self.controller.get_group()
        self.show_subframe('readfile', ColumnDataFileFrame,
                           group=dgroup.raw,
                           last_array_sel=self.last_array_sel,
                           _larch=self.larch,
                           read_ok_cb=partial(self.onRead_OK,
                                              overwrite=True))

    def onLoadFitResult(self, event=None):
        pass
        # print("onLoadFitResult??")
        # self.nb.SetSelection(1)
        # self.nb_panels[1].onLoadFitResult(event=event)

    def onReadDialog(self, event=None):
        dlg = wx.FileDialog(self, message="Read Data File",
                            defaultDir=os.getcwd(),
                            wildcard=FILE_WILDCARDS,
                            style=wx.FD_OPEN|wx.FD_MULTIPLE)
        self.paths2read = []
        if dlg.ShowModal() == wx.ID_OK:
            self.paths2read = dlg.GetPaths()
        dlg.Destroy()

        if len(self.paths2read) < 1:
            return

        path = self.paths2read.pop(0)
        path = path.replace('\\', '/')
        do_read = True
        if path in self.controller.file_groups:
            do_read = (wx.ID_YES == Popup(self,
                                          "Re-read file '%s'?" % path,
                                          'Re-read file?'))
        if do_read:
            self.onRead(path)

    def onRead(self, path):
        filedir, filename = os.path.split(os.path.abspath(path))
        if self.controller.get_config('chdir_on_fileopen'):
            os.chdir(filedir)
            self.controller.set_workdir()

        # check for athena projects
        if is_athena_project(path):
            kwargs = dict(filename=path,
                          _larch=self.controller.larch,
                          read_ok_cb=self.onReadAthenaProject_OK)
            self.show_subframe('athena_import', AthenaImporter, **kwargs)
        else:
            kwargs = dict(filename=path,
                          _larch=self.larch_buffer.larchshell,
                          last_array_sel = self.last_array_sel,
                          read_ok_cb=self.onRead_OK)

            self.show_subframe('readfile', ColumnDataFileFrame, **kwargs)

    def onReadAthenaProject_OK(self, path, namelist):
        """read groups from a list of groups from an athena project file"""
        self.larch.eval("_prj = read_athena('{path:s}', do_fft=False, do_bkg=False)".format(path=path))
        dgroup = None
        script = "{group:s} = extract_athenagroup(_prj.{prjgroup:s})"

        cur_panel = self.nb.GetCurrentPage()
        cur_panel.skip_plotting = True
        for gname in namelist:
            cur_panel.skip_plotting = (gname == namelist[-1])
            this = getattr(self.larch.symtable._prj, gname)
            gid = str(getattr(this, 'athena_id', gname))
            if self.larch.symtable.has_group(gid):
                count, prefix = 0, gname[:3]
                while count < 1e7 and self.larch.symtable.has_group(gid):
                    gid = prefix + make_hashkey(length=7)
                    count += 1

            self.larch.eval(script.format(group=gid, prjgroup=gname))
            dgroup = self.install_group(gid, gname, process=True, plot=False)
        self.larch.eval("del _prj")
        cur_panel.skip_plotting = False

        if len(namelist) > 0:
            gname = self.controller.file_groups[namelist[0]]
            self.ShowFile(groupname=gname, process=True, plot=True)
        self.write_message("read %d datasets from %s" % (len(namelist), path))

    def onRead_OK(self, script, path, groupname=None, filename=None,
                  array_sel=None, overwrite=False):
        """ called when column data has been selected and is ready to be used
        overwrite: whether to overwrite the current datagroup, as when
        editing a datagroup
        """
        if groupname is None:
            return
        abort_read = False
        filedir, real_filename = os.path.split(path)
        if filename is None:
            filename = real_filename
        if not overwrite and hasattr(self.larch.symtable, groupname):
            groupname = file2groupname(real_filename, symtable=self.larch.symtable)

        if abort_read:
            return

        self.larch.eval(script.format(group=groupname, path=path))
        if array_sel is not None:
            self.last_array_sel = array_sel
        self.install_group(groupname, filename, overwrite=overwrite)

        # check if rebin is needed
        thisgroup = getattr(self.larch.symtable, groupname)

        do_rebin = False
        if thisgroup.datatype == 'xas':
            try:
                en = thisgroup.energy
            except:
                do_rebin = True
                en = thisgroup.energy = thisgroup.xdat
            # test for rebinning:
            #  too many data points
            #  unsorted energy data or data in angle
            #  too fine a step size at the end of the data range
            if (len(en) > 1200 or
                any(np.diff(en) < 0) or
                ((max(en)-min(en)) > 300 and
                 (np.diff(en[-50:]).mean() < 0.75))):
                msg = """This dataset may need to be rebinned.
                Rebin now?"""
                dlg = wx.MessageDialog(self, msg, 'Warning',
                                       wx.YES | wx.NO )
                do_rebin = (wx.ID_YES == dlg.ShowModal())
                dlg.Destroy()

        for path in self.paths2read:
            path = path.replace('\\', '/')
            filedir, real_filename = os.path.split(path)
            gname = file2groupname(real_filename, symtable=self.larch.symtable)
            self.larch.eval(script.format(group=gname, path=path))
            self.install_group(gname, real_filename, overwrite=overwrite)

        self.write_message("read %s" % (real_filename))

        if do_rebin:
            RebinDataDialog(self, self.controller).Show()

    def install_group(self, groupname, filename, overwrite=False,
                      process=True, rebin=False, plot=True):
        """add groupname / filename to list of available data groups"""

        thisgroup = getattr(self.larch.symtable, groupname)

        datatype = getattr(thisgroup, 'datatype', 'raw')
        # file /group may already exist in list
        if filename in self.controller.file_groups and not overwrite:
            fbase, i = filename, 0
            while i < 10000 and filename in self.controller.file_groups:
                filename = "%s_%d" % (fbase, i)
                i += 1

        cmds = ["%s.groupname = '%s'" % (groupname, groupname),
               "%s.filename = '%s'" % (groupname, filename)]

        self.larch.eval('\n'.join(cmds))

        self.controller.filelist.Append(filename)
        self.controller.file_groups[filename] = groupname

        self.nb.SetSelection(0)
        self.ShowFile(groupname=groupname, process=process, plot=plot)
        self.controller.filelist.SetStringSelection(filename)
        return thisgroup


class XASViewer(wx.App, wx.lib.mixins.inspection.InspectionMixin):
    def __init__(self, filename=None, **kws):
        self.filename = filename
        wx.App.__init__(self, **kws)

    def run(self):
        self.MainLoop()

    def createApp(self):
        frame = XASFrame(filename=self.filename)
        self.SetTopWindow(frame)

    def OnInit(self):
        self.createApp()
        return True

def xas_viewer(**kws):
    XASViewer(**kws)



#!/usr/bin/env python
"""
  Larch shell utilities:
    more()
    ls()
    cwd()
    cd()
"""

import os
import sys
from copy import copy, deepcopy
from glob import glob

def _copy(obj):
    """copy an object"""
    return copy(obj)

def _deepcopy(obj):
    """deep copy an object"""
    return deepcopy(obj)

def _parent(name, _larch=None):
    "return parent group name of an object"
    return _larch.symtable._lookup(name)

def _ls(directory='.'):
    """return a list of files in the current directory,
    optionally using '*' to match file names

    Returns
    -------
    a : list of strings
       matching file names

    Examples
    --------
    to list all files::

        larch> ls('.')

    to list all files that end with '.xdi'::

        larch> ls('*.xdi')


    """
    directory.strip()
    if len(directory) == 0:
        arg = '.'
    if os.path.isdir(directory):
        ret = os.listdir(directory)
    else:
        ret = glob(directory)
    if sys.platform == 'win32':
        for i in range(len(ret)):
            ret[i] = ret[i].replace('\\','/')
    return ret

def _cwd():
    "return current working directory"
    ret = os.getcwd()
    if sys.platform == 'win32':
        ret = ret.replace('\\','/')
    return ret

def _cd(name):
    """change directory to specified directory"""
    name = name.strip()
    if name:
        os.chdir(name)

    ret = os.getcwd()
    if sys.platform == 'win32':
        ret = ret.replace('\\','/')
    return ret

def _mkdir(name, mode=0o777):
    """create directory (and any intermediate subdirectories

    Options:
    --------
      mode   permission mask to use for creating directory (default=0777)
    """
    return os.makedirs(name, mode=mode)

def show_more(text, filename=None, writer=None,
              pagelength=30, prefix='', _larch=None):
    """show lines of text in the style of more """
    txt = text[:]
    if isinstance(txt, str):
        txt = txt.split('\n')
    if len(txt) <1:
        return
    prompt = '== hit return for more, q to quit'
    ps = "%s (%%.2f%%%%) == " % prompt
    if filename:
        ps = "%s (%%.2f%%%%  of %s) == " % (prompt, filename)

    if writer is None:
        writer = sys.stdout

    i = 0
    for i in range(len(txt)):
        if txt[i].endswith('\n'):
            _larch.writer.write("%s%s" % (prefix, txt[i]))
        else:
            writer.write("%s%s\n" % (prefix, txt[i]))
        i = i + 1
        if i % pagelength == 0:
            try:
                x = input(ps %  (100.*i/len(txt)))
                if x in ('q','Q'): return
            except KeyboardInterrupt:
                writer.write("\n")
                return

def _more(fname, pagelength=32, _larch=None):
    """list file contents:
    > more('file.txt')
by default, the file is shown 32 lines at a time.
You can specify the number of lines to show at a time
with the  pagelength option:
    > more('file.txt', pagelength=10)
    """
    output = _larch.writer.write
    if not os.path.exists(fname):
        output("File '%s' not found.\n" % fname)
        return

    elif not os.path.isfile(fname):
        output("'%s' not a file.\n" % fname)
        return

    try:
        text = open(fname, 'r').readlines()
    except IOError:
        output("cannot open file: %s\n" % fname)
        return

    show_more(text, filename=fname, _larch=_larch,
              pagelength=pagelength)



"""
This module defines a device-independent MultiChannel Analyzer (MCA) class.

Authors/Modifications:
----------------------
* Mark Rivers, GSECARS
* Modified for Tdl, tpt
* modified and simplified for Larch, M Newville
"""
import numpy as np
from larch import Group, isgroup

from xraydb import xray_line, xray_edge, material_mu
from ..math import interp
from .deadtime import calc_icr, correction_factor
from .roi import ROI


def isLarchMCAGroup(grp):
    """tests whether variable holds a valid Larch MCAGroup"""
    return isgroup(grp, 'energy',  'counts', 'rois')

class Environment:
    """
    The "environment" or related parameters for a detector.  These might include
    things like motor positions, temperature, anything that describes the
    experiment.

    Attributes:
    -----------
    * name         # A string name of this parameter, e.g. "13IDD:m1"
    * description  # A string description of this parameter, e.g. "X stage"
    * value        # A string value of this parameter,  e.g. "14.223"
    """
    def __init__(self, desc='', addr='', val=''):
        self.addr  = str(addr).strip()
        self.val   = str(val).strip()
        self.desc  = str(desc).strip()
        if self.desc.startswith('(') and self.desc.endswith(')'):
            self.desc = self.desc[1:-1]

    def __repr__(self):
        return "Environment(desc='%s', val='%s', addr='%s')" % (self.desc,
                                                                self.val,
                                                                self.addr)


class MCA(Group):
    """
    MultiChannel Analyzer (MCA) class

    Attributes:
    -----------
    * self.name        = 'mca'  # Name of the mca object
    * self.nchans      = 2048   # number of mca channels
    * self.counts      = None   # MCA data
    * self.pileup      = None   # predicted pileup

    # Counting parameters
    * self.start_time   = ''    # Start time and date, a string
    * self.live_time    = 0.    # Elapsed live time in seconds
    * self.real_time    = 0.    # Elapsed real time in seconds
    * self.read_time    = 0.    # Time that the MCA was last read in seconds
    * self.total_counts = 0.    # Total counts between the preset start and stop channels
    * self.input_counts = 0.    # Actual total input counts (eg given by detector software)
                                # Note total_counts and input counts ARE NOT time normalized
                                #
    * self.tau          = -1.0  # Factor for deadtime/detector saturation calculations, ie
                                #     ocr = icr * exp(-icr*tau)
    * self.icr_calc     = -1.0  # Calculated input count rate from above expression
    * self.dt_factor   = 1.0    # deadtime correction factor based on icr,ocr,lt,rt
                                # data_corrected = data * dt_factor
                                #
    # Calibration parameters
    * self.offset       = 0.    # Offset
    * self.slope        = 1.0   # Slope
    * self.quad         = 0.    # Quadratic

    Notes:
    ------
    If input_counts are read from the data file then there is no need for tau
    values (and hence icr_calc is not used) - ie we assume the detector
    provides the icr value.

    The application of corrections uses the following logic:
       if tau > 0 this will be used in the correction factor calculation
       if tau = 0 then we assume ocr = icr in the correction factor calculation
       if tau < 0 (or None):
       if input_counts > 0 this will be used for icr in the factor calculation
          if input_counts <= 0 we assume ocr = icr in the correction factor calculation

    Energy calibration is based on the following:
        energy = offset + slope*channel + quad*channel**2

    """
    ###############################################################################
    def __init__(self, counts=None, nchans=2048, start_time='',
                 offset=0, slope=0, quad=0, name='mca', dt_factor=1,
                 real_time=0, live_time=0, input_counts=0, tau=0, **kws):

        self.name    = name
        self.nchans  = nchans
        self.environ = []
        self.rois    = []
        self.counts  = counts

        # Calibration parameters
        self.offset       = offset # Offset
        self.slope        = slope  # Slope
        self.quad         = quad   # Quadratic

        # Counting parameters
        self.start_time   = start_time
        self.real_time    = real_time  # Elapsed real time in seconds (requested counting time)
        self.live_time    = live_time  # Elapsed live time in seconds (time detector is live)
        self.input_counts = input_counts # Actual total input counts (eg given by detector software)
                                  # Note total_counts and input counts ARE NOT time normalized
                                  #
        self.total_counts = 0.    # Total counts between the preset start and stop channels
        self.tau          = tau   # Factor for deadtime/detector saturation calculations, ie
                                  #     ocr = icr * exp(-icr*tau)
        # Calculated correction values
        self.icr_calc    = -1.0  # Calculated input count rate from above expression
        # corrected_counts = counts * dt_factor
        self.bgr = None
        self.dt_factor   = float(dt_factor)
        if counts is not None:
            self.nchans      = len(counts)
            self.total_counts =  counts.sum()
        self.incident_energy = None
        self.get_energy()
        self._calc_correction()
        Group.__init__(self)

    def __repr__(self):
        form = "<MCA %s, nchans=%d, counts=%d, realtime=%.1f>"
        return form % (self.name, self.nchans, self.total_counts, self.real_time)

    def add_roi(self, name='', left=0, right=0, bgr_width=3,
                counts=None, sort=True):
        """add an ROI"""
        name = name.strip()
        roi = ROI(name=name, left=left, right=right,
                  bgr_width=bgr_width, counts=counts)

        rnames = [r.name.lower() for r in self.rois]
        if name.lower() in rnames:
            iroi = rnames.index(name.lower())
            self.rois[iroi] = roi
        else:
            self.rois.append(roi)
        if sort:
            self.rois.sort()

    def get_roi_counts(self, roi, net=False):
        """get counts for an roi"""
        thisroi = None
        if isinstance(roi, ROI):
            thisroi = roi
        elif isinstance(roi, int) and roi < len(self.rois):
            thisroi = self.rois[roi]
        elif isinstance(roi, str):
            rnames = [r.name.lower() for r in self.rois]
            for iroi, nam in enumerate(rnames):
                if nam.startswith(roi.lower()):
                    thisroi = self.rois[iroi]
                    break
        if thisroi is None:
            return None
        return thisroi.get_counts(self.counts, net=net)

    def add_environ(self, desc='', val='', addr=''):
        """add an Environment setting"""
        if len(desc) > 0 and len(val) > 0:
            self.environ.append(Environment(desc=desc, val=val, addr=addr))

    def predict_pileup(self, scale=None):
        """
        predict pileup for a spectrum, save to 'pileup' attribute
        """
        en = self.energy
        npts = len(en)
        counts = self.counts.astype(int)*1.0
        pileup = 1.e-8*np.convolve(counts, counts, 'full')[:npts]
        ex = en[0] + np.arange(len(pileup))*(en[1] - en[0])
        if scale is None:
            npts = len(en)
            nhalf = int(npts/2) + 1
            nmost = int(7*npts/8.0) - 1
            scale = self.counts[nhalf:].sum()/ pileup[nhalf:nmost].sum()
        self.pileup = interp(ex, scale*pileup, self.energy, kind='cubic')
        self.pileup_scale = scale

    def predict_escape(self, det='Si', scale=1.0):
        """
        predict detector escape for a spectrum, save to 'escape' attribute

        X-rays penetrate a depth 1/mu(material, energy) and the
        detector fluorescence escapes from that depth as
            exp(-mu(material, KaEnergy)*thickness)
        with a fluorecence yield of the material

        """
        fluor_en = xray_line(det, 'Ka').energy/1000.0
        edge = xray_edge(det, 'K')

        # here we use the 1/e depth for the emission
        # and the 1/e depth of the incident beam:
        # the detector fluorescence can escape on either side
        mu_emit  = material_mu(det, fluor_en*1000)
        mu_input = material_mu(det, self.energy*1000)
        escape   = 0.5*scale*edge.fyield * np.exp(-mu_emit / (2*mu_input))
        escape[np.where(self.energy*1000 < edge.energy)] = 0.0
        escape *= interp(self.energy - fluor_en, self.counts*1.0,
                         self.energy, kind='cubic')
        self.escape = escape

    def update_correction(self, tau=None):
        """
        Update the deadtime correction

        if tau == None just recompute,
        otherwise assign a new tau and recompute
        """
        if tau is not None:
            self.tau = tau
        self._calc_correction()

    def _calc_correction(self):
        """
        if self.tau > 0 this will be used in the correction factor calculation
        if self.tau = 0 then we assume ocr = icr in the correction factor calculation,
                      ie only lt correction
                     (note deadtime.calc_icr handles above two conditions)
        if self.tau < 0 (or None):
           if input_counts > 0  this will be used for icr in the factor calculation
           if input_counts <= 0 we assume ocr = icr in the correction factor calculation,
                                ie only lt correction
        """
        if self.live_time <= 0 or self.real_time <= 0:
            self.dt_factor  = 1.0
            return

        if self.total_counts > 0:
            ocr = self.total_counts / self.live_time
        else:
            ocr = None

        if self.tau >= 0:
            icr = calc_icr(ocr,self.tau)
            if icr is None:
                icr = 0
            self.icr_calc = icr
        elif self.input_counts > 0:
            icr = self.input_counts / self.live_time
        else:
            icr = ocr = None
        self.dt_factor  = correction_factor(self.real_time, self.live_time,
                                             icr=icr,  ocr=ocr)
        if self.dt_factor <= 0:
            print( "Error computing counts correction factor --> setting to 1")
            self.dt_factor = 1.0

    ########################################################################
    def get_counts(self, correct=True):
        """
        Returns the counts array from the MCA

        Note if correct == True the corrected counts is returned. However,
        this does not (re)compute the correction factor, therefore, make
        sure the correction factor is up to date before requesting
        corrected counts...
        """
        if correct:
            return  (self.dt_factor * self.counts).astype(np.int)
        else:
            return self.counts

    def get_energy(self):
        """
        Returns array of energy for each channel in the MCA spectra.
        """
        chans = np.arange(self.nchans, dtype=np.float)
        self.energy = self.offset +  chans * (self.slope + chans * self.quad)
        return self.energy

    def save_columnfile(self, filename, headerlines=None):
        "write summed counts to simple ASCII column file for mca counts"
        f = open(filename, "w+")
        f.write("#XRF counts for %s\n" % self.name)
        if headerlines is not None:
            for i in headerlines:
                f.write("#%s\n" % i)
        f.write("#\n")
        f.write("#EnergyCalib.offset = %.9g \n" % self.offset)
        f.write("#EnergyCalib.slope = %.9g \n" % self.slope)
        f.write("#EnergyCalib.quad  = %.9g \n" % self.quad)
        f.write("#Acquire.RealTime  = %.9g \n" % self.real_time)
        f.write("#Acquire.LiveTime  = %.9g \n" % self.live_time)
        roiform = "#ROI_%i '%s': [%i, %i]\n"
        for i, r in enumerate(self.rois):
            f.write(roiform % (i+1, r.name, r.left, r.right))

        f.write("#-----------------------------------------\n")
        f.write("#    energy       counts     log_counts\n")

        for e, d in zip(self.energy, self.counts):
            dlog = 0.
            if  d > 0: dlog = np.log10(max(d, 1))
            f.write(" %10.4f  %12i  %12.6g\n" % (e, d, dlog))
        f.write("\n")
        f.close()

    def dump_mcafile(self):
        """return text of mca file, not writing to disk, as for dump/load"""
        b = ['VERSION:    3.1', 'ELEMENTS:   1',
             'DATE:       %s' % self.start_time,
             'CHANNELS:   %d' % len(self.counts),
             'REAL_TIME:  %f' % self.real_time,
             'LIVE_TIME:  %f' % self.live_time,
             'CAL_OFFSET: %e' % self.offset,
             'CAL_SLOPE:  %e' % self.slope,
             'CAL_QUAD:   %e' % self.quad,
             'ROIS:       %d' % len(self.rois)]

        # ROIS
        for i, roi in enumerate(self.rois):
            b.extend(['ROI_%i_LEFT:  %d' % (i, roi.left),
                      'ROI_%i_RIGHT: %d' % (i, roi.right),
                      'ROI_%i_LABEL: %s &' % (i, roi.name)])

        # environment
        for e in self.environ:
            b.append('ENVIRONMENT: %s="%s" (%s)' % (e.addr, e.val, e.desc))

        # data
        b.append('DATA: ')
        for d in self.counts:
            b.append(" %d" % d)
        b.append('')
        return '\n'.join(b)

    def save_mcafile(self, filename):
        """write MCA file

        Parameters:
        -----------
        * filename: output file name
        """
        with open(filename, 'w') as fh:
            fh.write(self.dump_mcafile())

def create_mca(counts=None, nchans=2048, offset=0, slope=0, quad=0,
               name='mca', start_time='', real_time=0, live_time=0,
               dt_factor=1, input_counts=0, tau=0, **kws):

    """create an MCA object, containing an XRF (or similar) spectrum

     Parameters:
     ------------
      counts:   counts array
      nchans:   # channels

     Returns:
     ----------
      an MCA object

    """
    return MCA(counts=counts, nchans=nchans, name=name,
               start_time=start_time, offset=offset, slope=slope,
               quad=quad, dt_factor=dt_factor, real_time=real_time,
               live_time=live_time, input_counts=input_counts,
               tau=tau, **kws)



'''
This module defines a functions necessary for tomography calculations.

Authors/Modifications:
----------------------
* Margaret Koker, koker@cars.uchicago.edu
'''

import logging

logger = logging.getLogger(__name__)

import numpy as np
from scipy.optimize import leastsq, minimize

HAS_tomopy = False
try:
    import tomopy
    HAS_tomopy = True
except ImportError:
    pass

# GLOBAL VARIABLES
TOMOPY_ALG = ['gridrec', 'art', 'bart', 'mlem', 'osem', 'ospml_hybrid',
              'ospml_quad', 'pml_hybrid', 'pml_quad', 'sirt' ]

TOMOPY_FILT = ['shepp', 'ramlak', 'butterworth','parzen', 'cosine', 'hann',
               'hamming', 'None']

PIXEL_TRIM = 10

def reshape_sinogram(A,x=[],omega=[]):

    ## == INPUTS ==
    ## A              :    array from .get_roimap()
    ## x              :    x array for checking shape of A
    ## omega          :    omega array for checking shape of A
    ##
    ## == RETURNS ==
    ## A              :    A in shape/format needed for tomopy reconstruction
    ## sinogram_order :  flag/argument for tomopy reconstruction (shape dependent)

    A = np.array(A)
    if len(x) == len(omega):
        print('''Cannot reorder sinogram based on length of positional
                 arrays when same length. Acceptable orders:
                 sinogram_order = False : sino = [ 2th   , slice, X ]
                 sinogram_order = True  : sino = [ slice , 2th  , X ]''')
        return A,False
    if len(x) < 1 or len(omega) < 1:
        return A,False
    if len(A.shape) != 3:
       if len(A.shape) == 2:
           A = A.reshape(1,A.shape[0],A.shape[1])

    if len(A.shape) == 3:
        if len(x) == A.shape[0]:
             A = np.einsum('kij->ijk', A)
        if len(x) == A.shape[1]:
             A = np.einsum('ikj->ijk', A)
    sinogram_order = len(omega) == A.shape[1]

    return A,sinogram_order

def trim_sinogram(sino,x,omega,pixel_trim=None):

    if pixel_trim is None: pixel_trim = PIXEL_TRIM

    if len(omega) == sino.shape[-1]:
        omega = omega[pixel_trim:-1*(pixel_trim+1)]
    elif len(x) == sino.shape[-1]:
        x = x[pixel_trim:-1*(pixel_trim+1)]

    sino = sino[:,pixel_trim:-1*(pixel_trim+1)]

    return sino,x,omega

def find_tomo_center(sino, omega, center=None, sinogram_order=True):

    xmax = sino.shape[0]
    if sinogram_order:
        xmax = sino.shape[1]
    if center is None:
        center = xmax/2.0

    # init center to scale recon
    rec = tomopy.recon(sino, omega, center=center, sinogram_order=sinogram_order,
                       algorithm='gridrec', filter_name='shepp')
    rec = tomopy.circ_mask(rec, axis=0)

    # tomopy score, tweaked slightly
    rmin, rmax = rec.min(), rec.max()
    rmin  -= 0.5*(rmax-rmin)
    rmax  += 0.5*(rmax-rmin)
    out = minimize(_center_resid_negent, center,
                   args=(sino, omega, rmin, rmax, sinogram_order),
                   method='Nelder-Mead', tol=0.5)
    cen = out.x
    # if cen > 0  and cen < xmax:
    #    out = minimize(_center_resid_blur, cen,
    #                   args=(sino, omega, rmin, rmax, sinogram_order),
    #                   method='Nelder-Mead', tol=0.5)
    #    cen = out.x
    return cen

def _center_resid_negent(center, sino, omega, rmin, rmax, sinogram_order=True):
    """
    Cost function used for the ``find_center`` routine.
    """
    _, nang, nx = sino.shape
    if center < 1:
        return 10*(1-center)
    if center > nx-2:
        return 10*(center-nx+2)
    n1 = int(nx/4.0)
    n2 = int(3*nx/4.0)
    rec = tomopy.recon(sino, omega, center,algorithm='gridrec',
                       sinogram_order=sinogram_order)
    rec = tomopy.circ_mask(rec, axis=0)[:, n1:n2, n1:n2]
    hist, e = np.histogram(rec, bins=64, range=[rmin, rmax])
    hist = hist/rec.size
    score = -np.dot(hist, np.log(1.e-12+hist))
    logger.info("negent center = %.4f  %.4f" % (center, score))
    return score

def _center_resid_blur(center, sino, omega, rmin, rmax, sinogram_order=True):
    """
    Cost function used for the ``find_center`` routine.
    """
    rec = tomopy.recon(sino, omega, center,
                       sinogram_order=sinogram_order,
                       algorithm='gridrec', filter_name='shepp')
    rec = tomopy.circ_mask(rec, axis=0)
    score = -((rec - rec.mean())**2).sum()
    logger.info("blur center = %.4f  %.4f" % (center, score))
    return score

def tomo_reconstruction(sino, omega, algorithm='gridrec',
                        filter_name='shepp', num_iter=1, center=None,
                        refine_center=False, sinogram_order=True):
    '''
    INPUT ->  sino : slice, 2th, x OR 2th, slice, x (with flag sinogram_order=True/False)
    OUTPUT -> tomo : slice, x, y
    '''
    if center is None:
        center = sino.shape[1]/2.

    if refine_center:
        print(">> Refine Center start>> ", center, sinogram_order)
        # center = tomopy.find_center(sino, np.radians(omega), init=center,
        #                            ind=0, tol=0.5, sinogram_order=sinogram_order)

        center = find_tomo_center(sino, np.radians(omega), center=center,
                                  sinogram_order=sinogram_order)
        print(">> Refine Center done>> ", center, sinogram_order)
    algorithm = algorithm.lower()
    recon_kws = {}
    if algorithm.startswith('gridr'):
        recon_kws['filter_name'] = filter_name
    else:
        recon_kws['num_iter'] = num_iter
    tomo = tomopy.recon(sino, np.radians(omega), algorithm=algorithm,
                        center=center, sinogram_order=sinogram_order, **recon_kws)
    return center, tomo



#!/usr/bin/env python
from __future__ import print_function

import os
import sys
from time import time, sleep, ctime
import signal
import socket
from subprocess import Popen
from threading import Thread
from argparse import ArgumentParser, RawDescriptionHelpFormatter

from xmlrpc.server import SimpleXMLRPCServer
from xmlrpc.client import ServerProxy

from .interpreter import Interpreter
from .site_config import uname
from .utils.jsonutils import encode4js
from .utils import uname

try:
    import psutil
    HAS_PSUTIL = True
except ImportError:
    HAS_PSUTIL = False

NOT_IN_USE, CONNECTED, NOT_LARCHSERVER = range(3)
POLL_TIME = 0.50

"""Notes:
   0.  test server with HOST/PORT, report status (CREATED, ALREADY_RUNNING, FAILED).
   1.  prompt to kill a running server on HOST/PORT, preferably giving a
       'last used by {APPNAME} with {PROCESS_ID} at {DATETIME}'
   2.  launch server on next unused PORT on HOST, increment by 1 to 100, report status.
   3.  connect to running server on HOST/PORT.
   4.  have each client set a keepalive time (that is,
       'die after having no activity for X seconds') for each server (default=3*24*3600.0).
"""

def test_server(host='localhost', port=4966):
    """Test for a Larch server on host and port

    Arguments
      host (str): host name ['localhost']
      port (int): port number [4966]

    Returns
      integer status number:
          0    Not in use.
          1    Connected, valid Larch server
          2    In use, but not a valid Larch server
    """
    server = ServerProxy('http://%s:%d' % (host, port))
    try:
        methods = server.system.listMethods()
    except socket.error:
        return NOT_IN_USE

    # verify that this is a valid larch server
    if len(methods) < 5 or 'larch' not in methods:
        return NOT_LARCHSERVER
    ret = ''
    try:
        ret = server.get_rawdata('_sys.config.user_larchdir')
    except:
        return NOT_LARCHSERVER
    if len(ret) < 1:
        return NOT_LARCHSERVER

    return CONNECTED


def get_next_port(host='localhost', port=4966, nmax=100):
    """Return next available port for a Larch server on host

    Arguments
      host (str): host name ['localhost']
      port (int): starting port number [4966]
      nmax (int): maximum number to try [100]

    Returns
      integer: next unused port number or None in nmax exceeded.
    """
    # special case for localhost:
    # use psutil to find next unused port
    if host.lower() == 'localhost':
        if HAS_PSUTIL and uname == 'win':
            available = [True]*nmax
            try:
                conns = psutil.net_connections()
            except:
                conns = []
            if len(conns) > 0:
                for conn in conns:
                    ptest = conn.laddr[1] - port
                    if ptest >= 0 and ptest < nmax:
                        available[ptest] = False
            for index, status in enumerate(available):
                if status:
                    return port+index
        # now test with brute attempt to open the socket:
        for index in range(nmax):
            ptest = port + index
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
            success = False
            try:
                sock.bind(('', ptest))
                success = True
            except socket.error:
                pass
            finally:
                sock.close()
            if success:
                return ptest

    # for remote servers or if the above did not work, need to test ports
    for index in range(nmax):
        ptest = port + index
        if NOT_IN_USE == test_server(host=host, port=ptest):
            return ptest
    return None

class LarchServer(SimpleXMLRPCServer):
    def __init__(self, host='localhost', port=4966,
                 logRequests=False, allow_none=True,
                 keepalive_time=3*24*3600):
        self.out_buffer = []

        self.larch = Interpreter(writer=self)
        self.larch.input.prompt = ''
        self.larch.input.prompt2 = ''
        self.larch.run_init_scripts()

        self.larch('_sys.client = group(keepalive_time=%f)' % keepalive_time)
        self.larch('_sys.wx = group(wxapp=None)')
        _sys = self.larch.symtable._sys
        _sys.color_exceptions = False
        _sys.client.last_event = int(time())
        _sys.client.pid_server = int(os.getpid())
        _sys.client.app = 'unknown'
        _sys.client.pid = 0
        _sys.client.user = 'unknown'
        _sys.client.machine = socket.getfqdn()

        self.client = self.larch.symtable._sys.client
        self.port = port
        SimpleXMLRPCServer.__init__(self, (host, port),
                                    logRequests=logRequests,
                                    allow_none=allow_none)

        self.register_introspection_functions()
        self.register_function(self.larch_exec, 'larch')

        for method in ('ls', 'chdir', 'cd', 'cwd', 'shutdown',
                        'set_keepalive_time', 'set_client_info',
                        'get_client_info', 'get_data', 'get_rawdata',
                        'get_messages', 'len_messages'):
            self.register_function(getattr(self, method), method)

        # sys.stdout = self
        self.finished = False
        signal.signal(signal.SIGINT, self.signal_handler)
        self.activity_thread = Thread(target=self.check_activity)

    def write(self, text, **kws):
        if text is None:
            text = ''
        self.out_buffer.append(str(text))

    def flush(self):
        pass

    def set_keepalive_time(self, keepalive_time):
        """set keepalive time
        the server will self destruct after keepalive_time of inactivity

        Arguments:
            keepalive_time (number): time in seconds

        """
        self.larch("_sys.client.keepalive_time = %f" % keepalive_time)

    def set_client_info(self, key, value):
        """set client info

        Arguments:
            key (str): category
            value (str): value to use

        Notes:
            the key can actually be any string but include by convention:
               app      application name
               user     user name
               machine  machine name
               pid      process id
        """
        self.larch("_sys.client.%s = '%s'" % (key, value))

    def get_client_info(self):
        """get client info:
        returns json dictionary of client information
        """
        out = {'port': self.port}
        client = self.larch.symtable._sys.client
        for attr in dir(client):
            out[attr] = getattr(client, attr)
        return encode4js(out)

    def get_messages(self):
        """get (and clear) all output messages (say, from "print()")
        """
        out = "".join(self.out_buffer)
        self.out_buffer = []
        return out

    def len_messages(self):
        "length of message buffer"
        return len(self.out_buffer)

    def ls(self, dir_name):
        """list contents of a directory: """
        return os.listdir(dir_name)

    def chdir(self, dir_name):
        """change directory"""
        return os.chdir(dir_name)

    def cd(self, dir_name):
        """change directory"""
        return os.chdir(dir_name)

    def cwd(self):
        """change directory"""
        ret = os.getcwd()
        if uname == 'win':
            ret = ret.replace('\\','/')
        return ret

    def signal_handler(self, sig=0, frame=None):
        self.kill()

    def kill(self):
        """handle alarm signal, generated by signal.alarm(t)"""
        sleep(POLL_TIME)
        self.shutdown()
        self.server_close()

    def shutdown(self):
        "shutdown LarchServer"
        self.finished = True
        if self.activity_thread.is_alive():
            self.activity_thread.join(POLL_TIME)
        return 1

    def check_activity(self):
        while not self.finished:
            sleep(POLL_TIME)
            # print("Tick ", time()- (self.client.keepalive_time + self.client.last_event))
            if time() > (self.client.keepalive_time + self.client.last_event):
                t = Thread(target=self.kill)
                t.start()
                break

    def larch_exec(self, text):
        "execute larch command"
        text = text.strip()
        if text in ('quit', 'exit', 'EOF'):
            self.shutdown()
        else:
            ret = self.larch.eval(text, lineno=0)
            if ret is not None:
                self.write(repr(ret))
            self.client.last_event = time()
            self.flush()
        return 1

    def get_rawdata(self, expr):
        "return non-json encoded data for a larch expression"
        return self.larch.eval(expr)

    def get_data(self, expr):
        "return json encoded data for a larch expression"
        self.larch('_sys.client.last_event = %i' % time())
        return encode4js(self.larch.eval(expr))

    def run(self):
        """run server until times out"""
        self.activity_thread.start()
        while not self.finished:
            try:
                self.handle_request()
            except:
                break

def spawn_server(port=4966, wait=True, timeout=30):
    """
    start a new process for a LarchServer on selected port,
    optionally waiting to confirm connection
    """
    topdir = sys.exec_prefix
    pyexe = os.path.join(topdir, 'bin', 'python')
    bindir = 'bin'
    if uname.startswith('win'):
            bindir = 'Scripts'
            pyexe = pyexe + '.exe'

    args = [pyexe, os.path.join(topdir, bindir, 'larch'),
            '-r', '-p', '%d' % port]
    pipe = Popen(args)
    if wait:
        t0 = time()
        while time() - t0 < timeout:
            sleep(POLL_TIME)
            if CONNECTED == test_server(port=port):
                break
    return pipe


###
def larch_server_cli():
    """command-line program to control larch XMLRPC server"""
    command_desc = """
command must be one of the following:
  start       start server on specified port
  stop        stop server on specified port
  restart     restart server on specified port
  next        start server on next avaialable port (see also '-n' option)
  status      print a short status message: whether server< is running on port
  report      print a multi-line status report
"""

    parser = ArgumentParser(description='run larch XML-RPC server',
                            formatter_class=RawDescriptionHelpFormatter,
                            epilog=command_desc)

    parser.add_argument("-p", "--port", dest="port", default='4966',
                        help="port number for remote server [4966]")

    parser.add_argument("-n", "--next", dest="next", action="store_true",
                        default=False,
                        help="show next available port, but do not start [False]")

    parser.add_argument("-q", "--quiet", dest="quiet", action="store_true",
                        default=False, help="suppress messaages [False]")

    parser.add_argument("command", nargs='?',  help="server command ['status']")

    args = parser.parse_args()


    port = int(args.port)
    command = args.command or 'status'
    command = command.lower()

    def smsg(port, txt):
        if not args.quiet:
            print('larch_server port=%i: %s' % (port, txt))


    if args.next:
        port = get_next_port(port=port)
        print(port)
        sys.exit(0)

    server_state = test_server(port=port)

    if command == 'start':
        if server_state == CONNECTED:
            smsg(port, 'already running')
        elif server_state == NOT_IN_USE:
            spawn_server(port=port)
            smsg(port, 'started')
        else:
            smsg(port, 'port is in use, cannot start')

    elif command == 'stop':
        if server_state == CONNECTED:
            ServerProxy('http://localhost:%d' % (port)).shutdown()
            smsg(port, 'stopped')

    elif command == 'next':
        port = get_next_port(port=port)
        spawn_server(port=port)
        smsg(port, 'started')

    elif command == 'restart':
        if server_state == CONNECTED:
            ServerProxy('http://localhost:%d' % (port)).shutdown()
            sleep(POLL_TIME)
        spawn_server(port=port)

    elif command == 'status':
        if server_state == CONNECTED:
            smsg(port, 'running')
            sys.exit(0)
        elif server_state == NOT_IN_USE:
            smsg(port, 'not running')
            sys.exit(1)
        else:
            smsg(port, 'port is in use by non-larch server')
    elif command == 'report':
        if server_state == CONNECTED:
            s = ServerProxy('http://localhost:%d' % (port))
            info = s.get_client_info()
            last_event = info.get('last_event', 0)
            last_used = ctime(last_event)
            serverid  = int(info.get('pid_server', 0))
            serverport= int(info.get('port', 0))
            procid    = int(info.get('pid', 0))
            appname   = info.get('app',     'unknown')
            machname  = info.get('machine', 'unknown')
            username  = info.get('user',    'unknown')
            keepalive_time = info.get('keepalive_time', -1)
            keepalive_time += (last_event - time())
            keepalive_units = 'seconds'
            if keepalive_time > 150:
                keepalive_time = round(keepalive_time/60.0)
                keepalive_units = 'minutes'
            if keepalive_time > 150:
                keepalive_time = round(keepalive_time/60.0)
                keepalive_units = 'hours'

            print('larch_server report:')
            print('   Server Port Number  = %s' % serverport)
            print('   Server Process ID   = %s' % serverid)
            print('   Server Last Used    = %s' % last_used)
            print('   Server will expire in %d %s if not used.' % (keepalive_time,
                                                                 keepalive_units))
            print('   Client Machine Name = %s' % machname)
            print('   Client Process ID   = %s' % str(procid))
            print('   Client Application  = %s' % appname)
            print('   Client User Name    = %s' % username)

        elif server_state == NOT_IN_USE:
            smsg(port, 'not running')
            sys.exit(1)
        else:
            smsg(port, 'port is in use by non-larch server')

    else:
        print("larch_server: unknown command '%s'. Try -h" % command)


if __name__ == '__main__':
    spawn_server(port=4966)



#!/usr/bin/env python
"""
Larch Filling:  stolen and hacked from PyCrust's Filling module

Filling is the gui tree control through which a user can navigate
the local namespace or any object."""

from __future__ import print_function

__author__ = "Patrick K. O'Brien <pobrien@orbtech.com>"
__cvsid__ = "$Id: filling.py 37633 2006-02-18 21:40:57Z RD $"
__revision__ = "$Revision: 37633 $"

import sys
import wx
import numpy
import wx.html as html
import types

from wx.py import dispatcher
from wx.py import editwindow

import inspect
from functools import partial

from wx.py import introspect
from larch.symboltable import SymbolTable, Group
from larch.larchlib import Procedure
from wxutils  import Button, pack, is_wxPhoenix


VERSION = '0.9.5(Larch)'

COMMONTYPES = (int, float, complex, str, bool, dict, list, tuple, numpy.ndarray)

H5TYPES = ()
try:
    import h5py
    H5TYPES = (h5py.File, h5py.Group, h5py.Dataset)
except ImportError:
    pass

TYPE_HELPS = {}
for t in COMMONTYPES:
    TYPE_HELPS[t] = 'help on %s' % t

IGNORE_TYPE = []
DOCTYPES = ('BuiltinFunctionType', 'BuiltinMethodType', 'ClassType',
            'FunctionType', 'GeneratorType', 'InstanceType',
            'LambdaType', 'MethodType', 'ModuleType',
            'UnboundMethodType', 'method-wrapper')

def rst2html(text):
    return "<br>".join(text.split('\n'))


def call_signature(obj):
    """try to get call signature for callable object"""
    fname = obj.__name__


    if isinstance(obj, partial):
        obj = obj.func

    argspec = None
    if hasattr(obj, '_larchfunc_'):
        obj = obj._larchfunc_

    argspec = inspect.getfullargspec(obj)
    keywords = argspec.varkw

    fargs = []
    ioff = len(argspec.args) - len(argspec.defaults)
    for iarg, arg in enumerate(argspec.args):
        if arg == '_larch':
            continue
        if iarg < ioff:
            fargs.append(arg)
        else:
            fargs.append("%s=%s" % (arg, repr(argspec.defaults[iarg-ioff])))
    if keywords is not None:
        fargs.append("**%s" % keywords)

    out = "%s(%s)" % (fname, ', '.join(fargs))
    maxlen = 71
    if len(out) > maxlen:
        o  = []
        while len(out) > maxlen:
            ecomm = maxlen - out[maxlen-1::-1].find(',')
            o.append(out[:ecomm])
            out = " "*(len(fname)+1) + out[ecomm:].strip()
        if len(out)  > 0:
            o.append(out)
        out = '\n'.join(o)
    return out

class FillingTree(wx.TreeCtrl):
    """FillingTree based on TreeCtrl."""

    name = 'Larch Filling Tree'
    revision = __revision__

    def __init__(self, parent, id=-1, pos=wx.DefaultPosition,
                 size=wx.DefaultSize, style=wx.TR_DEFAULT_STYLE,
                 rootObject=None, rootLabel=None, rootIsNamespace=False,
                 static=False):
        """Create FillingTree instance."""
        wx.TreeCtrl.__init__(self, parent, id, pos, size, style)
        self.rootIsNamespace = rootIsNamespace
        self.rootLabel = rootLabel
        self.static = static
        self.item = None
        self.root = None
        if is_wxPhoenix:
            self.GetPyData = self.GetItemData

        self.setRootObject(rootObject)

    def setRootObject(self, rootObject=None):
        self.rootObject = rootObject
        if self.rootObject is None:
            return
        if not self.rootLabel:
            self.rootLabel = 'Larch Data'

        rootData = rootObject
        if not is_wxPhoenix:
            rootData = wx.TreeItemData(rootData)
        self.item = self.root = self.AddRoot(self.rootLabel, -1, -1,  rootData)

        self.SetItemHasChildren(self.root,  self.objHasChildren(self.rootObject))
        self.Bind(wx.EVT_TREE_ITEM_EXPANDING, self.OnItemExpanding, id=self.GetId())
        self.Bind(wx.EVT_TREE_ITEM_COLLAPSED, self.OnItemCollapsed, id=self.GetId())
        self.Bind(wx.EVT_TREE_SEL_CHANGED, self.OnSelChanged, id=self.GetId())
        # self.Bind(wx.EVT_TREE_ITEM_ACTIVATED, self.OnItemActivated, id=self.GetId())
        self.Bind(wx.EVT_RIGHT_DOWN, self.OnSelChanged, id=self.GetId() )
        if not self.static:
            dispatcher.connect(receiver=self.push, signal='Interpreter.push')

    def push(self, command, more):
        """Receiver for Interpreter.push signal."""
        self.display()

    def OnItemExpanding(self, event):
        """Add children to the item."""
        item = event.GetItem()
        if self.IsExpanded(item):
            return
        self.addChildren(item)
        self.SelectItem(item)

    def OnItemCollapsed(self, event):
        """Remove all children from the item."""
        item = event.GetItem()

    def OnSelChanged(self, event):
        """Display information about the item."""
        if hasattr(event, 'GetItem'):
            self.item = event.GetItem()
        self.display()

    def OnItemActivated(self, event):
        """Launch a DirFrame."""
        item = event.GetItem()
        text = self.getFullName(item)
        obj = self.GetPyData(item)
        frame = FillingFrame(parent=self, size=(500, 500),
                             rootObject=obj,
                             rootLabel=text, rootIsNamespace=False)
        frame.Show()

    def objHasChildren(self, obj):
        """Return true if object has children."""
        children = self.objGetChildren(obj)
        if isinstance(children, dict):
            return len(children) > 0
        else:
            return False

    def objGetChildren(self, obj):
        """Return dictionary with attributes or contents of object."""
        otype = type(obj)
        d = {}
        if (obj is None or obj is False or obj is True):
            return d
        self.ntop = 0
        if isinstance(obj, SymbolTable) or isinstance(obj, Group):
            d = obj._members()
        if isinstance(obj, COMMONTYPES):
            d = obj
        elif isinstance(obj, h5py.Group):
            try:
                for key, val in obj.items():
                    d[key] = val
            except (AttributeError, ValueError):
                pass
        elif isinstance(obj, h5py.Dataset):
            d = obj
        elif isinstance(obj, (list, tuple)):
            for n in range(len(obj)):
                key = '[' + str(n) + ']'
                d[key] = obj[n]
        elif (not isinstance(obj, wx.Object)
              and not hasattr(obj, '__call__')):
            d = self.GetAttr(obj)
        return d

    def GetAttr(self, obj):
        out = {}
        for key in dir(obj):
            if not ((key.startswith('__') and key.endswith('__')) or
                    key.startswith('_SymbolTable') or
                    key == '_main'):
                try:
                    out[key] = getattr(obj, key)
                except:
                    out[key] = key
        return out

    def addChildren(self, item):
        self.DeleteChildren(item)
        obj = self.GetPyData(item)
        children = self.objGetChildren(obj)
        if not children:
            return
        try:
            keys = children.keys()
        except:
            return
        # keys.sort(lambda x, y: cmp(str(x).lower(), str(y).lower()))
        for key in sorted(keys):
            itemtext = str(key)
            # Show string dictionary items with single quotes, except
            # for the first level of items, if they represent a
            # namespace.
            if (isinstance(obj, dict) and isinstance(key, str) and
                (item != self.root or
                 (item == self.root and not self.rootIsNamespace))):
                itemtext = repr(key)
            child = data = children[key]
            if not is_wxPhoenix:
                data = wx.TreeItemData(child)
            branch = self.AppendItem(parent=item, text=itemtext, data=data)
            self.SetItemHasChildren(branch, self.objHasChildren(child))

    def display(self):
        item = self.item
        if not item:
            return
        obj = self.GetPyData(item)
        if self.IsExpanded(item):
            self.addChildren(item)
        self.setText('')

        if wx.Platform == '__WXMSW__':
            if obj is None: # Windows bug fix.
                return
        self.SetItemHasChildren(item, self.objHasChildren(obj))
        otype = type(obj)
        text = []
        fullname = self.getFullName(item)
        if fullname is not None:
            text.append("%s\n" % fullname)

        needs_doc = False
        if isinstance(obj, COMMONTYPES):
            text.append('Type: %s' % otype.__name__)
            text.append('Value = %s' % repr(obj))

        elif isinstance(obj, Group):
            text.append('Group: %s ' % obj.__name__)
            gdoc = getattr(obj, '__doc__', None)
            if gdoc is None: gdoc = Group.__doc__
            text.append(gdoc)
        elif hasattr(obj, '__call__'):
            text.append('Function: %s' % obj)
            try:
                text.append("\n%s" % call_signature(obj))
            except:
                pass
            needs_doc = True
        else:
            text.append('Type: %s' % str(otype))
            text.append('Value = %s' % repr(obj))
        text.append('\n')
        if needs_doc:
            try:
                doclines = obj.__doc__.strip().split('\n')
            except:
                doclines = ['No documentation found']
            indent = 0
            for dline in doclines:
                if len(dline.strip()) > 0:
                    indent = dline.index(dline.strip())
                    break
            for d in doclines:
                text.append(d[indent:])
        text.append('\n')
        self.setText('\n'.join(text))

    def getFullName(self, item, part=''):
        """Return a syntactically proper name for item."""
        try:
            name = self.GetItemText(item)
        except:
            return None

        # return 'Could not get item name: %s' % repr(item)
        parent = None
        obj = None
        if item != self.root:
            parent = self.GetItemParent(item)
            obj = self.GetPyData(item)
        # Apply dictionary syntax to dictionary items, except the root
        # and first level children of a namepace.
        if ((isinstance(obj, dict) or hasattr(obj, 'keys')) and
            ((item != self.root and parent != self.root) or
             (parent == self.root and not self.rootIsNamespace))):
            name = '[' + name + ']'
        # Apply dot syntax to multipart names.
        if part:
            if part[0] == '[':
                name += part
            else:
                name += '.' + part
        # Repeat for everything but the root item
        # and first level children of a namespace.
        if (item != self.root and parent != self.root) \
        or (parent == self.root and not self.rootIsNamespace):
            name = self.getFullName(parent, part=name)
        return name

    def setText(self, text):
        """Display information about the current selection."""

        # This method will likely be replaced by the enclosing app to
        # do something more interesting, like write to a text control.
        print( text)

    def setStatusText(self, text):
        """Display status information."""

        # This method will likely be replaced by the enclosing app to
        # do something more interesting, like write to a status bar.
        print( text)


class FillingText(editwindow.EditWindow):
    """FillingText based on StyledTextCtrl."""

    name = 'Filling Text'
    revision = __revision__

    def __init__(self, parent, id=-1, pos=wx.DefaultPosition,
                 size=wx.DefaultSize, style=wx.CLIP_CHILDREN,
                 static=False):
        """Create FillingText instance."""
        editwindow.EditWindow.__init__(self, parent, id, pos, size, style)
        # Configure various defaults and user preferences.
        self.SetReadOnly(True)
        self.SetWrapMode(True)
        self.SetMarginWidth(1, 0)
        if not static:
            dispatcher.connect(receiver=self.push, signal='Interpreter.push')

    def push(self, command, more):
        """Receiver for Interpreter.push signal."""
        self.Refresh()

    def SetText(self, *args, **kwds):
        # print("Text Set Text ", args)
        self.SetReadOnly(False)
        editwindow.EditWindow.SetText(self, *args, **kwds)
        self.SetReadOnly(True)


class FillingRST(html.HtmlWindow):
    """FillingText based on Rest doc string!"""

    name = 'Filling Restructured Text'

    def __init__(self, parent, id=-1, pos=wx.DefaultPosition,
                 size=wx.DefaultSize, style=wx.NO_FULL_REPAINT_ON_RESIZE,
                 static=False):
        """Create FillingRST instance."""
        html.HtmlWindow.__init__(self, parent, id, style=wx.NO_FULL_REPAINT_ON_RESIZE)

        # Configure various defaults and user preferences.
        if not static:
            dispatcher.connect(receiver=self.push, signal='Interpreter.push')

    def push(self, command, more):
        """Receiver for Interpreter.push signal."""
        pass

    def SetText(self, text='', **kwds):
        # html = ['<html><body>',rst2html(text), '</body></html>']
        self.SetPage(rst2html(text))



class Filling(wx.SplitterWindow):
    """Filling based on wxSplitterWindow."""

    name = 'Filling'
    revision = __revision__

    def __init__(self, parent, id=-1, pos=wx.DefaultPosition,
                 size=wx.DefaultSize, style=wx.SP_3D|wx.SP_LIVE_UPDATE,
                 name='Filling Window', rootObject=None,
                 rootLabel=None, rootIsNamespace=False, static=False):
        """Create a Filling instance."""
        wx.SplitterWindow.__init__(self, parent, id, pos, size, style, name)
        self.tree = FillingTree(parent=self, rootObject=rootObject,
                                rootLabel=rootLabel,
                                rootIsNamespace=rootIsNamespace,
                                static=static)

        self.text = FillingText(parent=self, static=static)
        self.SplitVertically(self.tree, self.text, 200)
        self.SetMinimumPaneSize(100)

        # Override the filling so that descriptions go to FillingText.
        self.tree.setText = self.text.SetText

        # Display the root item.
        if self.tree.root is not None:
            self.tree.SelectItem(self.tree.root)
            self.tree.display()

        self.Bind(wx.EVT_SPLITTER_SASH_POS_CHANGED, self.OnChanged)

    def SetRootObject(self, rootObject=None):
        self.tree.setRootObject(rootObject)
        if self.tree.root is not None:
            self.tree.SelectItem(self.tree.root)
            wx.CallAfter(self.tree.display)

    def OnChanged(self, event):
        #this is important: do not evaluate this event=>
        # otherwise, splitterwindow behaves strange
        #event.Skip()
        pass

    def onRefresh(self, evt=None):
        """ refesh data tree, preserving current selection"""
        root = self.tree.GetRootItem()
        this = self.tree.GetFocusedItem()
        parents = [self.tree.GetItemText(this)]
        while True:
            try:
                this = self.tree.GetItemParent(this)
                if this == root:
                    break
                parents.append(self.tree.GetItemText(this))
            except:
                break
        self.tree.Collapse(root)
        self.tree.Expand(root)
        node = root

        while len(parents) > 0:
            name = parents.pop()
            node = self.get_node_by_name(node, name)
            if node is not None:
                self.tree.Expand(node)

        try:
            self.tree.Expand(node)
            self.tree.SelectItem(node)
        except:
            pass

    def get_node_by_name(self, node, name):
        if node is None:
            node = self.tree.GetRootItem()
        item, cookie = self.tree.GetFirstChild(node)
        if item.IsOk() and self.tree.GetItemText(item) == name:
            return item

        nodecount = self.tree.GetChildrenCount(node)
        while nodecount > 1:
            nodecount -= 1
            item, cookie = self.tree.GetNextChild(node, cookie)
            if not item.IsOk() or self.tree.GetItemText(item) == name:
                return item

    def ShowNode(self, name):
        """show node by name"""
        root = self.tree.GetRootItem()
        self.tree.Collapse(root)
        self.tree.Expand(root)
        node = root
        parts = name.split('.')
        parts.reverse()
        while len(parts) > 0:
            name = parts.pop()
            node = self.get_node_by_name(node, name)
            if node is not None:
                self.tree.Expand(node)

        try:
            self.tree.Expand(node)
            self.tree.SelectItem(node)
        except:
            pass

    def LoadSettings(self, config):
        pos = config.ReadInt('Sash/FillingPos', 200)
        wx.FutureCall(250, self.SetSashPosition, pos)
        zoom = config.ReadInt('View/Zoom/Filling', -99)
        if zoom != -99:
            self.text.SetZoom(zoom)

    def SaveSettings(self, config):
        config.WriteInt('Sash/FillingPos', self.GetSashPosition())
        config.WriteInt('View/Zoom/Filling', self.text.GetZoom())


class FillingFrame(wx.Frame):
    """Frame containing the namespace tree component."""
    name = 'Filling Frame'
    revision = __revision__
    def __init__(self, parent=None, id=-1, title='Larch Data Tree',
                 pos=wx.DefaultPosition, size=(600, 400),
                 style=wx.DEFAULT_FRAME_STYLE, rootObject=None,
                 rootLabel=None, rootIsNamespace=False, static=False):
        """Create FillingFrame instance."""
        wx.Frame.__init__(self, parent, id, title, pos, size, style)
        intro = 'Larch Data Tree'
        self.CreateStatusBar()
        self.SetStatusText(intro)
        self.filling = Filling(parent=self,
                               rootObject=rootObject,
                               rootLabel=rootLabel,
                               rootIsNamespace=rootIsNamespace,
                               static=static)
        # Override so that status messages go to the status bar.
        self.filling.tree.setStatusText = self.SetStatusText


class App(wx.App):
    """PyFilling standalone application."""
    def OnInit(self):
        wx.InitAllImageHandlers()
        self.fillingFrame = FillingFrame()
        self.fillingFrame.Show(True)
        self.SetTopWindow(self.fillingFrame)
        return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-

r"""
Generic discrete convolution

Description
-----------

 This is a manual (not optimized!) implementation of discrete 1D
 convolution intended for spectroscopy analysis. The difference with
 commonly used methods is the possibility to adapt the convolution
 kernel for each convolution point, e.g. change the FWHM of the
 Gaussian kernel as a function of the energy scale.

Resources
---------

.. [WPconv] <http://en.wikipedia.org/wiki/Convolution#Discrete_convolution>
.. [Fisher] <http://homepages.inf.ed.ac.uk/rbf/HIPR2/convolve.htm>
.. [GP1202] <http://glowingpython.blogspot.fr/2012/02/convolution-with-numpy.html>

TODO
----
- [] get_ene_index: substitute with more elegant
     'np.argmin(np.abs(ene-(x)))'
- [] atan_gamma_fdmnes: define atan_gamma as in FDMNES
"""

from __future__ import print_function

__author__ = "Mauro Rovezzi"
__email__ = "mauro.rovezzi@gmail.com"
__credits__ = ""
__license__ = "BSD license <http://opensource.org/licenses/BSD-3-Clause>"
__organization__ = "European Synchrotron Radiation Facility"
__owner__ = "Mauro Rovezzi"
__year__ = "2013-2014"
__version__ = "0.0.4"
__status__ = "Alpha"
__date__ = "Jan 2014"

DEBUG = 0

import os, sys
import subprocess
from optparse import OptionParser
from datetime import date
from string import Template
import numpy as np

from .lineshapes import gaussian, lorentzian

def get_ene_index(ene, cen, hwhm):
    """ returns the min/max indexes for array ene at (cen-hwhm) and (cen+hwhm)
    very similar to index_of in larch
    """
    try:
        if ((cen-hwhm) <= min(ene)):
            ene_imin = 0
        else:
            ene_imin = max(np.where(ene < (cen-hwhm))[0])
        if ((cen+hwhm) >= max(ene)):
            ene_imax = (len(e)-1)
        else:
            ene_imax = min(np.where(ene > (cen+hwhm))[0])
        return ene_imin, ene_imax
    except:
        print('index not found for {0} +/- {1}'.format(cen, hwhm))
        return None, None

def lin_gamma(ene, fwhm=1.0, linbroad=None):
    """returns constant or linear energy-dependent broadening

    Parameters
    ----------
    ene : energy array in eV
    fwhm : first full width at half maximum in eV
    linbroad : list of 3-elements giving
               'second full width at half maximum'
               'energy starting point of the linear increase'
               'energy ending point of the linear increase'
    """
    w = np.ones_like(ene)
    if linbroad is None:
        return w*fwhm
    else:
        try:
            fwhm2 = linbroad[0]
            e1 = linbroad[1]
            e2 = linbroad[2]
        except:
            raise ValueError('wrong format for linbroad')
        for en, ee in enumerate(ene):
            if ee < e1:
                w[en] *= fwhm
            elif ee <= e2:
                wlin = (fwhm + (ee - e1) * (fwhm2 - fwhm) / (e2 - e1))
                w[en] *= wlin
            elif ee >= e2:
                w[en] *= fwhm2
        return w

def atan_gamma(ene, gamma_hole, gamma_max=15., e0=0, eslope=1.):
    r"""returns arctangent-like broadening, $\Gamma(E)$

    ..math

    \Gamma(E) = \Gamma_{hole} + \Gamma_{max} * ( \arctan( \frac{E-E_{0}}{E_{slope}} ) / \pi + 1/2) )
    """
    if (eslope == 0):
        print('Warning: eslope cannot be zero, using default value of 1')
        eslope = 1.
    return gamma_hole + gamma_max * ( ( np.arctan( (ene - e0) / eslope ) / np.pi ) + 0.5 )

def conv(e, mu, kernel='gaussian', fwhm_e=None, efermi=None):
    """ linear broadening

    Parameters
    ----------
    e : x-axis (energy)
    mu : f(x) to convolve with g(x) kernel, mu(energy)
    kernel : convolution kernel, g(x)
             'gaussian'
             'lorentzian'
    fwhm_e: the full width half maximum in eV for the kernel
            broadening. It is an array of size 'e' with constants or
            an energy-dependent values determined by a function as
            'lin_gamma()' or 'atan_gamma()'
    """
    f = np.copy(mu)
    z = np.zeros_like(f)
    if efermi is not None:
        #ief = index_nearest(e, efermi)
        ief = np.argmin(np.abs(e-efermi))
        f[0:ief] *= 0
    if e.shape != fwhm_e.shape:
        print("Error: 'fwhm_e' does not have the same shape of 'e'")
        return 0
    # linar fit upper part of the spectrum to avoid border effects
    # polyfit => pf
    lpf = len(e)/2.
    cpf = np.polyfit(e[-lpf:], f[-lpf:], 1)
    fpf = np.poly1d(cpf)
    # extend upper energy border to 3*fhwm_e[-1]
    estep = (e[-1] - e[-2])
    eup = np.append(e, np.arange(e[-1]+estep, e[-1]+3*fwhm_e[-1], estep))
    for n in range(len(f)):
        #from now on I change e with eup
        eimin, eimax = get_ene_index(eup, eup[n], 1.5*fwhm_e[n])
        if ((eimin is None) or (eimax is None)):
            if DEBUG:
                raise IndexError('e[{0}]'.format(n))
        if len(range(eimin, eimax)) % 2 == 0:
            kx = eup[eimin:eimax+1] # odd range centered at the convolution point
        else:
            kx = eup[eimin:eimax]
        ### kernel ###
        hwhm = fwhm_e[n]/2.0
        if ('gauss' in kernel.lower()):
            ky = gaussian(kx, center=eup[n], sigma=hwhm)
        elif ('lor' in kernel.lower()):
            ky = lorentzian(kx, center=eup[n], sigma=hwhm)
        else:
            raise ValueError("convolution kernel '{0}' not implemented".format(kernel))
        ky = ky/ky.sum() # normalize
        zn = 0
        lk = len(kx)
        for mf, mg in zip( range(-(lk/2), (lk/2)+1), range(lk) ):
            if ( ((n+mf) >= 0) and ((n+mf) < len(f)) ):
                zn += f[n+mf] * ky[mg]
            elif ((n+mf) >= 0):
                zn += fpf(eup[n+mf]) * ky[mg]
        z[n] = zn
    return z

def glinbroad(e, mu, fwhm_e=None, efermi=None):
    """ gaussian linear convolution in Larch """
    return conv(e, mu, kernel='gaussian', fwhm_e=fwhm_e, efermi=efermi)
glinbroad.__doc__ = conv.__doc__

### CONVOLUTION WITH FDMNES VIA SYSTEM CALL ###
class FdmnesConv(object):
    """ Performs convolution with FDMNES within Python """
    def __init__(self, opts=None, calcroot=None, fn_in=None, fn_out=None):
        if opts is None:
            self.opts = dict(creator = 'FDMNES toolbox',
                             today = date.today(),
                             calcroot = calcroot,
                             fn_in = fn_in,
                             fn_out = fn_out,
                             fn_ext = 'txt',
                             estart_sel = '',
                             estart = '-20.',
                             efermi_sel = '',
                             efermi = '-5.36',
                             spin = '',
                             core_sel = '!',
                             core = '!',
                             hole_sel = '',
                             hole = '0.5',
                             conv_const = '!',
                             conv_sel = '',
                             ecent = '25.0',
                             elarg = '20.0',
                             gamma_max = '10.0',
                             gamma_type = 'Gamma_fix',
                             gauss_sel = '',
                             gaussian = '0.9')
        else:
            self.opts = opts
        if calcroot is not None:
            self.opts['calcroot'] = calcroot
            self.opts['fn_in'] = '{}.{}'.format(calcroot,
                                                self.opts['fn_ext'])
            self.opts['fn_out'] = '{}_conv{}.{}'.format(calcroot,
                                                        self.opts['spin'],
                                                        self.opts['fn_ext'])
        if fn_in is not None:
            self.opts['calcroot'] = fn_in[:-4]
            self.opts['fn_in'] = fn_in
            self.opts['fn_out'] = '{}_conv{}.{}'.format(fn_in[:-4],
                                                        self.opts['spin'],
                                                        self.opts['fn_ext'])
        if fn_out is not None:
            self.opts['fn_out'] = fn_out
        # then check all options
        self.checkopts()

    def checkopts(self):
        if (self.opts['calcroot'] is None) or (self.opts['fn_in'] is None):
            raise NameError("missing 'calcroot' or 'fn_in'")
        if self.opts['estart'] == '!':
            self.opts['estart_sel'] = '!'
        if self.opts['efermi'] == '!':
            self.opts['efermi_sel'] = '!'
        if self.opts['spin'] == 'up':
            self.opts['core_sel'] = ''
            self.opts['core'] = '2 !spin up'
        elif self.opts['spin'] == 'down':
            self.opts['core_sel'] = ''
            self.opts['core'] = '1 !spin down'
        elif self.opts['spin'] == '':
            self.opts['core_sel'] = '!'
        elif self.opts['spin'] == 'both':
            raise NameError('spin="both" not implemented!')
        else:
            self.opts['spin'] = ''
            self.opts['core_sel'] = '!'
            self.opts['core'] = '!'
        if self.opts['hole'] == '!':
            self.opts['hole_sel'] = '!'
        if self.opts['conv_const'] == '!':
            self.opts['conv_sel'] = '!'
        else:
            self.opts['conv_sel'] = ''
        if self.opts['gamma_type'] == 'Gamma_fix':
            pass
        elif self.opts['gamma_type'] == 'Gamma_var':
            pass
        else:
            raise NameError('gamma_type="Gamma_fix"/"Gamma_var"')
        if self.opts['gaussian'] == '!':
            self.opts['gauss_sel'] = '!'
        else:
            self.opts['gauss_sel'] = ''
        # update the output file name
        self.opts['fn_out'] = '{}_conv{}.{}'.format(self.opts['calcroot'],
                                                    self.opts['spin'],
                                                    self.opts['fn_ext'])
    def setopt(self, opt, value):
        self.opts[opt] = value
        self.checkopts()

    def wfdmfile(self):
        """ write a simple fdmfile.txt to enable the convolution
        first makes a copy of previous fdmfile.txt if not already done """
        if os.path.exists('fdmfile.bak'):
            print('fdmfile.bak exists, good')
        else:
            subprocess.call('cp fdmfile.txt fdmfile.bak', shell=True)
            print('copied fdmfile.txt to fmdfile.bak')
        #
        s = Template('!fdmfile.txt automatically created by ${creator} on ${today} (for convolution)\n\
!--------------------------------------------------------------------!\n\
! Number of calculations\n\
 1\n\
! FOR CONVOLUTION STEP\n\
convfile.txt\n\
!--------------------------------------------------------------------!\n\
')
        outstr = s.substitute(self.opts)
        f = open('fdmfile.txt', 'w')
        f.write(outstr)
        f.close()

    def wconvfile(self):
        s = Template("""
!FDMNES convolution file\n\
!created by ${creator} on ${today}\n\
!
Calculation\n\
${fn_in}\n\
Conv_out\n\
${fn_out}\n\
${estart_sel}Estart\n\
${estart_sel}${estart}\n\
${efermi_sel}Efermi\n\
${efermi_sel}${efermi}\n\
${core_sel}Selec_core\n\
${core_sel}${core}\n\
${hole_sel}Gamma_hole\n\
${hole_sel}${hole}\n\
${conv_sel}Convolution\n\
${conv_sel}${ecent} ${elarg} ${gamma_max} !Ecent Elarg Gamma_max\n\
${conv_sel}${gamma_type}\n\
${gauss_sel}Gaussian\n\
${gauss_sel}${gaussian} !Gaussian conv for experimental res\n\
""")
        outstr = s.substitute(self.opts)
        f = open('convfile.txt', 'w')
        f.write(outstr)
        f.close()

    def run(self):
        """ runs fdmnes """
        self.wfdmfile()  # write fdmfile.txt
        self.wconvfile() # write convfile.txt
        try:
            subprocess.call('fdmnes', shell=True)
        except OSError:
            print("check 'fdmnes' executable exists!")



import numpy as np

from xraydb import xray_line, xray_edge, material_mu
from larch import  parse_group_args
from .xafsutils import set_xafsGroup
from .pre_edge import preedge

def fluo_corr(energy, mu, formula, elem, group=None, edge='K', line='Ka', anginp=45,
              angout=45, _larch=None, **pre_kws):
    """correct over-absorption (self-absorption) for fluorescene XAFS
    using the FLUO alogrithm of D. Haskel.

    Arguments
    ---------
      energy    array of energies
      mu        uncorrected fluorescence mu
      formula   string for sample stoichiometry
      elem      atomic symbol or Z of absorbing element
      group     output group [default None]
      edge      name of edge ('K', 'L3', ...) [default 'K']
      line      name of line ('K', 'Ka', 'La', ...) [default 'Ka']
      anginp    input angle in degrees  [default 45]
      angout    output angle in degrees  [default 45]

    Additional keywords will be passed to pre_edge(), which will be used
    to ensure consistent normalization.

    Returns
    --------
       None, writes `mu_corr` and `norm_corr` (normalized `mu_corr`)
       to output group.

    Notes
    -----
       Support First Argument Group convention, requiring group
       members 'energy' and 'mu'
    """
    energy, mu, group = parse_group_args(energy, members=('energy', 'mu'),
                                         defaults=(mu,), group=group,
                                         fcn_name='fluo_corr')
    # gather pre-edge options
    pre_opts = {'e0': None, 'nnorm': 1, 'nvict': 0,
                'pre1': None, 'pre2': -30,
                'norm1': 100, 'norm2': None}
    if hasattr(group, 'pre_edge_details'):
        uopts = getattr(group.pre_edge_details, 'call_args', {})
        for attr in pre_opts:
            if attr in uopts:
                pre_opts[attr] = uopts[attr]
    pre_opts.update(pre_kws)
    pre_opts['step'] = None
    pre_opts['nvict'] = 0

    # generate normalized mu for correction
    preinp   = preedge(energy, mu, **pre_opts)

    ang_corr = (np.sin(max(1.e-7, np.deg2rad(anginp))) /
                np.sin(max(1.e-7, np.deg2rad(angout))))

    # find edge energies and fluorescence line energy
    e_edge  = xray_edge(elem, edge).energy
    e_fluor = xray_line(elem, line).energy

    # calculate mu(E) for fluorescence energy, above, below edge

    muvals = material_mu(formula, np.array([e_fluor, e_edge-10.0,
                                            e_edge+10.0]), density=1)

    alpha   = (muvals[0]*ang_corr + muvals[1])/(muvals[2] - muvals[1])
    mu_corr = mu*alpha/(alpha + 1 - preinp['norm'])
    preout  = preedge(energy, mu_corr, **pre_opts)
    if group is not None:
        if _larch is not None:
            group = set_xafsGroup(group, _larch=_larch)
        group.mu_corr = mu_corr
        group.norm_corr = preout['norm']



#!/usr/bin/env python
"""Help topics for on-line documentation and help in larch

The data here contain python variable holding text that
are used as principle help topics.

M. Newville Univ of Chicago, T. Trainor, Univ of Alaska Fairbanks (2005,2006)

"""
##
## usage notes:
##    to add a new top level help topic, simply create a variable named '_h_%s' % topic
## where 'topic' is a unique (case-insensitive!) name for a help topic.
##
## when  help(topic)
##
_h_Shell = """Larch command-line shell help:

"""

_h_Help =  """
   ================================
   = Help options =
    help          # this help
    help topics   # list topics for additional help
    help name     # help on specific topic, variable, or procedure
   ================================
"""

_h_Overview = """
This is Larch (Matt Newville and Tom Trainor 2009).

Larch is a data processing language designed to be
  - easy to use for novices.
  - complete enough for intermediate to advanced data processing.
  - data-centric, so that arrays of data are easy to manage and use.
  - easily extendable with python.

Larch uses Numeric Python and SciPy for all its numerical functionality,
so that processing data in arrays is both easy and fast. There is a wide
range of 'advanced numeric capabilities' for Larch, including complex
numbers, FFTs and non-linear least squares minimization.

Larch Features:

==Simple Syntax:
  Variable and function names are simple, undecorated strings.  The
  syntax is deliberately close to Python, so that

==Data Types:

  Named variables can contain any of the following data types: ints,
  floats, complex, strings, lists, dictionaries (hashes), NumPy arrays,
  or any other python data types.

  Multi-dimensional data can be accessed either element-by-element or with
  more complicated "slices":
        larch> x   = 1             # int
        larch> y   = 7 + 2j        # complex number:  real + imag'j'
        larch> arr = arange(10)    # simple array
        larch> print arr
            [0 1 2 3 4 5 6 7 8 9 10]
        larch> print arr[3]
            3
        larch> print arr[2:8]
            [2 3 4 5 6 7]
        larch> str = 'here is a string'
        larch> print str[0:7]
            here is

==Full Featured, extensible:

  Many useful functions for numerical data processing are built in to
  Larch.  More importantly, it is easy to add more functions, either
  from Python or as Larch procedures.

==Namespaces and Groups:

  Each variable and function lives in Group, which can be nested much
  like a file directory structure -- Group.Subgroup.Variable.  These is
  no limit on the number of subgroups, within a group or the level of
  nesting of groups.  This allows you to organize data,
     print DataSet1.Energy
     print DataSet2.Energy

  and so on. To access a variable, you can always use the Full name:
     print Group1.SubgroupA.Variable

  If you ask for a name that is NOT a Full name, larch tries to look it up.

  It does this by always having a "local group" which is always looked in first.
  If

  Unqualified names (ie, a name without a 'group.' prefix) will use a default
  order to determine which Group to associate with a name:  There is always
  a "Default Data Group" and a "Default Function Group" each of which can be
  changed during a Larch session.  At startup, there are two pre-defined groups:
  "_builtin", which contains many built in functions and constants (sin and pi,
  for example), and "_main".  At startup, "_main" is empty, and is set as the
  "Default Data Group" and "Default Function Group".

  When creating or assigning a variable without an explicit group name, the default
  Data Group is used.  When creating or assigning a procedure without an explicit
  group name, the default Function Group is used.

  When using (say, on the right-hand side of a statement) a variable or function with
  an unqualified group name, the first matching name in the list
     (Default Data Group, Default Function Group, "_main", "_builtin")

  is used.  The Default Data Group can be changed with
        larch> datagroup('mydata')

  and the Default Function Group can be changed with
        larch> funcgroup('myfuncs')

  Assigning a fully qualified variable or function to a group that previously did
  not exist will create that group:
        larch> x = 1
        larch> group1.x = 3.3
        larch> print x
            1
        larch> datagroup('group1')
        larch> print x
            3.3
        larch> print _main.x, group1.x
            1 3.3
        larch> show groups
            Default Data Group = group1
            Default Function Group = _main
            All Groups:
                _main  _builtin  group1

==Functions and Commands.  Functions and commands have similiar name structure as
  data (group.func,group.cmd).
  The default group name for functions and commands is _main (functions and commands with
  this group name do not need full name qualification).
  Eg. you can also assign variables variable values with setvar() function:
        larch> setvar('v', 2.3, group='group1')
        larch> print group1.v

  The potential advantage here is that 'v' and 'group1' are string types,
  not literal variable names, so they can be computed or passed in a
  procedure.

==Clean syntax for programming:
  The syntax is "Python Inspired", with some notable exceptions.  Most
  importantly, indentation level is not significant, and blocks end with
  with explicit 'end' statements.  A typical for-block will look like this:

        for i in range(10):
            print i
        endfor

  There are also while blocks and if-elif-else blocks:
        n = 0
        while n<10:
           print ' n = ',n
           if n==3: print 'here is 3!!'
               n = n + 1
           endwhile

        if n > 10:
            print 'No'
        elif n > 5 and n < 8:
            print 'Maybe'
        else:
            print 'Yep'
        endif

  A design goal is that well-formed larch code should be very easy to
  translate into valid python (and vice versa).

==User-defined functions (aka procedures):
  User defined functions can be written in larch:
        def  myfunc(arg1, option='test'):
             'documentation string'
              print 'this is my funcition ', arg1
              print 'optional param = ', option
              if type(option) != 'string':
                  print 'option must be a string!!'
                  return False
              endif
              value = sqrt(arg1)
              x.tmp = value
              return value > 10.
        enddef

  which could be called as
        larch> ret = myfunc(3., option = 'xx')
        larch> print ret, x.tmp
          False 1.73205080757

==dofile() function:
  you can run a file of larch commands with the dofile() function:
        larch> dofile('myfile.larch')

==eval() function:
  you can construct a larch expression on the fly from strings and execute it:

        larch> eval("%s.value = %f" %  ['group', 10.2])
        larch> print group.value
          10.2

==read_ascii() function:
  you can read in ASCII column data files very easily.

        larch> read_ascii('my.dat', group='f')
          ['f', 'x', ''y']

  this read 2 columns (labeled 'x' and 'y') from the column file and
  created the array variables f.x and f.y.  Also created was f.titles
  to hold the titles in the data file (non-numeric content at the top of
  the file) and f.column_labels

==On-line help:
  well, this is in progress....

==Easy to add your python functions, including getting access to all
  the 'data groups' inside your python function.

==Differences between larch and Python (for python users):
  -  larch has many builtins and assumes Numerical data.
  -  larch has no tuples. Lists are used in their place.
  -  indentation does not matter. blocks are ended with 'end***'
     (if / endif , for /endfor, def/enddef)
  -  when in doubt, assignment makes a copy, and does not give a reference
  -  data types are NOT objects, and so have no methods.  You must use
     procedural approaches
       x = arange(100)
       reshape(x,[10,10])   instead of x.shape = (10,10)
"""

##Strings

_h_Strings = """
  Working with strings:

  strings are sequence of text characters. To specify a string, you enclose it in quotes,
  either single or double quotes:
      larch>  x = 'this is a string'
      larch>  x = "A string with a ' in it "

  There are a few special characters that can be included in strings by 'escaping' them with a
  backlash character "\\".  The most important of these are '\\n' to write a newline character,
  and '\\t' to write a tab character.  Other escaped characters are:
      \\n   newline
      \\r   carriage return
      \\f   form feed
      \\v   vertical tab
      \\b   backspace
      \\t   tab
      \\a   bell
      \\"   literal "
      \\'   literal '
      \\\   the backslach character itself.

  You can escape quote characters ('\\"' or '\\'') so that they do not mark the end of a
  string sequence:
      larch> x = 'a string\\'s string'
      larch> y = "\\"a string with a double quote\\", he said."


  ==Multi-line Strings with Triple Quotes

  Larch allows multi-line strings (that is strings that span lines).  One simple way to do this is
  to include a newline character ('\\n') character in the string, but this is not always sufficient.
  Another way is to use triple quotes (three quotes in a row: either ''' or  \"\"\") to enclose the
  string.  With this approach, the newlines of the enclosed string are preserved:

     larch> long_string = '''Here is a mult-line string
     ...> line #2
     ...> line #3 of the long string '''

     larch> print long_string
     Here is a mult-line string
     line #2
     line #3 of the long string

  As with single quote strings, you have the choice of using single or double quotes, and can
  use escaped character and escaped quotes.

  ==String Formatting

  It is often desirable to create strings from program data. To do this, you *format a string*.
  This is done by putting format codes in a string and supplying values to fill in.  Format
  codes use the '%' character to mark the formatting to do.

  Following normal conventions, a formatted string for a floating point number might look like this:

      larch> print "x = %8.4f" % sqrt(12)
      x =   3.4641

  The "%8.4f" tells the formatting to format a floating point number (the 'f') with 8 total numbers
  and 4 numbers after the decimal point.   A plain '%' is used to separate the format string and the
  value(s) to format.  Other format codes:

      ......

  ==Using a dictionary for string formatting

    Borrowing from Python, larch also allow you to format a string using a dictionary instead of a
    simple list.  This can be a big advantage when formatting many values or formatting data from
    complicated data structures.  In this method, you explicitly set the dictionary key to use by
    naming them between the '%' and the format code:

      larch> data  = {'name':'File 1', 'date':'Monday, April 3, 2006', 'x': 12.4}
      larch> print " %(name)s , x = %(x)f" % data
      File 1 , x = 12.4

"""

# dictionaries
_h_Dicts = """
  Working with dictionaries:

"""

# arrays
_h_Arrays = """
  Working with arrays:

"""

# lists
_h_Lists = """
  Working with lists:

"""

# data types
_h_DataTypes = """
   Supported data types:
"""

# data types
_h_Groups = """
Working with groups:
"""

# python
_h_Python = """
Working with python and python modules:
"""

#
_h_Control = """
   Help on Programming Control structures (conditionals, loops, etc)

   Program Control structures are important for controlling what parts of
   a program are run.  If you're familiar with other programming languages,
   the concepts and syntax here should be fairly straightforward.

   The basic concept here is to define blocks of code (that is multiple lines
   of code) to be run as a group.  This allows the block of code to be run
   only under some conditions, or to be run repeatedly with different values
   for some variables in the block.

   The syntax of larch uses a colon ':' and a small number of keywords to define
   the blocks of code:
       if x==0:
          print 'cannot divide by zero'
       else:
          x = 2 /x
       endif


 = If statements:

   if statements run a block of code only if some condition is met.  Since this
   is so common, there are a few variations allowed in larch.  First is the
   one line version:

       if x<0:  x = -x

   which will set x to -x (that is, this will make sure x>=0).  In general, the
   if statement looks like:
       if condition:

   where the ending ':' is important (to indicate where the condition end).
   The "condition" is a statement that is evaluated as a boolean value (either
   True or False -- all numeric values are True except 0.0).  The boolean
   operations 'and', 'or' , and 'not' to construct complex conditions:

       if x>0 and x<10:

       if (x>0 and (b>1 or c>1)):


   A warning: in many programming languages multiple conditions can be relied upon
   to only evaluate until the truth of the statement can be determined.  This is NOT
   the case in larch: the entire statement may be evaluated.

   The second version uses the same 'if condition:', but on a line by itself, followed
   by  multiple statements.  This version ends with 'endif' to indicate how far the
   block of code extends:
       if x < 0  :
           x = -x
           print ' x set to ',x
       endif

   Next, there's the 'if-else-endif' version which runs either one block or the
   other, and looks like this (note the
       if x<0:
           x = -x
           print ' x set to ',x
       else:
           print ' x was positive to begin with'
       endif

   The final and most complete variation uses 'elif' (pronounced "else if") to allow
   multiple conditions to be tested:

       if x<0:
           x = -x
           print ' x set to ',x
       elif x>10:
           x = 10
           print ' x was too big, set to ', x
       else:
           print ' x was OK to begin with'
       endif

   Multiple 'elif' blocks can be given, though all 'elif' statements must be
   after the 'if' statements, and an 'else' statement (if present) must be last.
   And the entire construct must end with an 'endif'

   In such constructs, the first block with a condition that is True, and no other
   blocks will be run.

 = For loops:

   for loops repeat a block of code, usually incrementing some value used in
   the loop.  Usually, a for loop runs a predictable number of times (see the
   section below on Break and Continue for when it does not!). A for loop
   looks like:

       for i in [1,2,3]:
           print i, i*i
       endfor

   The basic syntax for the first line is 'for <variable> in <list>:'  The list
   can be either a real list or a numerical array -- a very common approach is to
   use the range() function to generate a list of numbers:
       for i in range(10):
           print i, 10-i, i*(10-i)
           if i < 2: print ' i < 2!! '
       endfor

   The block is run repeatedly with the loop variable i set to each value in the
   list.

   There is also a 'one-line' version of the for loop:

       for i in [1,2,3]: call_some_function(i)


 = While loops:

   while loops are similar to for loops, repeatedly running a block of code
   as long as some condition is true:

      x = 1
      while x<4:
         print x
         x = x + 1
      endwhile

   prints
     1.0
     2.0
     3.0

   Beware that a while loop makes it easy to make an "infinite loop" (for example, if
   the value of x had not been increased in the block).

 = Break and Continue in For and While loops

   In both for loops and while loops, the block can be exited with a 'break' statement.


 = Try / Except blocks:

   Sometimes error happen, and it is desirable to run some block of code

       x = 0
       a = 2.
       try:
           y = a/ x
       except:
           print "OK,  that did not work so well!"
       endtry
"""

_h_If = ""
_h_For = ""
_h_While = ""
_h_Try = ""
_h_Builtin = ""
_h_Functions = ""
_h_Procedures = ""
_h_String_Formatting = "String Formatting"

def generate():
    topics = {'':_h_Help}
    for key,val in globals().items():
        if key.startswith('_h_'):
            name = key[3:].lower().replace('_',' ')
            topics[name] = val

    tnames = ['== Help topics ==']
    s = ''

    topnames = sorted(topics.keys())
    nx= 2 + sorted([len(i) for i in topnames])[-1]

    fmt = "%%%i.%is" % (nx,nx)
    for t in topnames:
        s = "%s %s" % (s,fmt%(t+(' '*nx)))
        if len(s) > 50:
            tnames.append(s)
            s = ''
    tnames.append(s)
    topics['topics'] = '\n'.join(tnames)

    return topics



#!/usr/bin/python
"""
support for netcdf file output files containing MCA spectra
from Epics Mapping Mode with XIA xMXAP electronics
"""
import numpy as np
import time
import sys
import os

try:
    import scipy.io.netcdf
    netcdf_open = scipy.io.netcdf.netcdf_file
except ImportError:
    raise ImportError('cannot find a netcdf module')

def aslong(d):
    """unravels and converts array of int16 (int) to int32 (long)"""
    # need to unravel the array!!!
    d = d.astype(np.int16).ravel()
    d.dtype = np.int32
    return d

class xMAPBufferHeader(object):
    def __init__(self,buff):
        self.tag0          = buff[0]  # Tag word 0
        self.tag1          = buff[1]  # Tag word 1
        self.headerSize    = buff[2]  #  Buffer header size
        #  Mapping mode (1=Full spectrum, 2=Multiple ROI, 3=List mode)
        self.mappingMode   = buff[3]
        self.runNumber     = buff[4]  # Run number
        # Sequential buffer number, low word first
        self.bufferNumber  = aslong(buff[5:7])[0]
        self.bufferID      = buff[7]  # 0=A, 1=B
        self.numPixels     = buff[8]  # Number of pixels in buffer
        # Starting pixel number, low word first
        self.startingPixel = aslong(buff[9:11])[0]
        self.moduleNumber  = buff[11]
        self.channelID     = np.array(buff[12:20]).reshape((4,2))
        self.channelSize   = buff[20:24]
        self.bufferErrors  = buff[24]
        self.userDefined   = buff[32:64]

class xMAPMCAPixelHeader(object):
    def __init__(self,buff):
        self.tag0        = buff[0]
        self.tag1        = buff[1]
        self.headerSize  = buff[2]
        # Mapping mode (1=Full spectrum, 2=Multiple ROI, 3=List mode)
        self.mappingMode = buff[3]
        self.pixelNumber = aslong(buff[4:6])[0]
        self.blockSize   = aslong(buff[6:8])[0]

class xMAPData(object):
    def __init__(self,npix,nmod,nchan):
        ndet = 4 * nmod
        self.firstPixel   = 0
        self.numPixels    = 0
        self.counts       = np.zeros((npix, ndet, nchan), dtype='i2')
        self.realTime     = np.zeros((npix, ndet), dtype='i8')
        self.liveTime     = np.zeros((npix, ndet), dtype='i8')
        self.inputCounts  = np.zeros((npix, ndet), dtype='i4')
        self.outputCounts = np.zeros((npix, ndet), dtype='i4')

CLOCKTICK = 0.320  # xmap clocktick = 320 ns

def read_xrf_netcdf(fname, npixels=None, verbose=False):
    # Reads a netCDF file created with the DXP xMAP driver
    # with the netCDF plugin buffers
    if verbose:
        print( ' reading ', fname)
    t0 = time.time()
    # read data from array_data variable of netcdf file
    read_ok = False
    fh = None
    try:
        fh = netcdf_open(fname,'r')
        read_ok = True
    except:
        time.sleep(0.010)
        try:
            fh = netcdf_open(fname,'r')
            read_ok = True
        except:
            pass

    if not read_ok or fh is None:
        if fh is not None:
            fh.close()
        return None

    array_data = fh.variables['array_data']
    t1 = time.time()

    # array_data will normally be 3d:
    #  shape = (narrays, nmodules, buffersize)
    # but nmodules and narrays could be 1, so that
    # array_data could be 1d or 2d.
    #
    # here we force the data to be 3d
    shape = array_data.shape
    if len(shape) == 1:
        array_data.shape = (1, 1, shape[0])
    elif len(shape) == 2:
        array_data.shape = (1, shape[0], shape[1])

    narrays, nmodules, buffersize = array_data.shape
    modpixs    = int(max(124, array_data[0, 0, 8]))
    npix_total = 0
    # real / live times are returned in microseconds.
    for array in range(narrays):
        for module in range(nmodules):
            d   = array_data[array,module, :]
            bh  = xMAPBufferHeader(d)
            #if verbose and array==0:
            dat = d[256:].reshape(modpixs, int((d.size-256)/modpixs ))

            npix = bh.numPixels
            if module == 0:
                npix_total += npix
                if array == 0:
                    # first time through, (array,module)=(0,0) we
                    # read mapping mode, set up how to slice the
                    # data, and build data arrays in xmapdat
                    mapmode = dat[0, 3]
                    if mapmode == 1:  # mapping, full spectra
                        nchans = d[20]
                        data_slice = slice(256, 8448)
                    elif mapmode == 2:  # ROI mode
                        # Note:  nchans = number of ROIS !!
                        nchans     = max(d[264:268])
                        data_slice = slice(64, 64+8*nchans)
                    xmapdat = xMAPData(narrays*modpixs, nmodules, nchans)
                    xmapdat.firstPixel = bh.startingPixel

            # acquistion times and i/o counts data are stored
            # as longs in locations 32:64
            t_times = aslong(dat[:npix, 32:64]).reshape(npix, 4, 4)
            p1 = npix_total - npix
            p2 = npix_total
            xmapdat.realTime[p1:p2, :]     = t_times[:, :, 0]
            xmapdat.liveTime[p1:p2, :]     = t_times[:, :, 1]
            xmapdat.inputCounts[p1:p2, :]  = t_times[:, :, 2]
            xmapdat.outputCounts[p1:p2, :] = t_times[:, :, 3]

            # the data, extracted as per data_slice and mapmode
            t_data = dat[:npix, data_slice]
            if mapmode == 2:
                t_data = aslong(t_data)
            xmapdat.counts[p1:p2, :, :] = t_data.reshape(npix, 4, nchans)

    t2 = time.time()
    xmapdat.numPixels = npix_total
    xmapdat.counts    = xmapdat.counts[:npix_total]
    xmapdat.realTime = CLOCKTICK * xmapdat.realTime[:npix_total]
    xmapdat.liveTime = CLOCKTICK * xmapdat.liveTime[:npix_total]
    xmapdat.inputCounts  = xmapdat.inputCounts[:npix_total]
    xmapdat.outputCounts = xmapdat.outputCounts[:npix_total]
    if verbose:
        print('   time to read file    = %5.1f ms' % ((t1-t0)*1000))
        print('   time to extract data = %5.1f ms' % ((t2-t1)*1000))
        print('   read %i pixels ' %  npix_total)
        print('   data shape:    ' ,  xmapdat.counts.shape)
    fh.close()
    return xmapdat

def test_read(fname):
    print( fname,  os.stat(fname))
    fd = read_xrf_netcdf(fname, verbose=True)
    print(fd.counts.shape)



"""
Basic Fitting Models for 1-D data, simplifying fits to many standard line shapes.

  usage:
  ------
  param_group = fit_peak(x, y, model, dy=None,
                         background='linear', form='linear')

  arguments:
  ---------
  x           array of values at which to calculate model
  y           array of values for model to try to match
  dy          array of values for uncertainty in y data to be matched.
  model       name of model to use.  One of (case insensitive)
                 'linear', 'quadratic', 'step', 'rectangle',
                 'exponential', 'gaussian', 'lorentzian', 'voigt'

  background  name of background model to use. One of (case insensitive)
                 None, 'constant', 'linear', or 'quadratic'
              this is ignored when model is 'linear' or 'quadratic'
  form        name of form to use for 'step' and 'rectangle' models.
              One of (case insensitive):
                 'linear', 'erf', or 'atan'
  output:
  -------
  param_group   Group with fit parameters, and
"""

import numpy as np
from scipy.special import gamma, gammaln, beta, betaln, erf, erfc, wofz
from lmfit import Parameter, Minimizer
from lmfit.model import Model

from lmfit.models import (update_param_vals, LinearModel, ConstantModel,
                          QuadraticModel, PolynomialModel, GaussianModel,
                          LorentzianModel, VoigtModel, PseudoVoigtModel,
                          MoffatModel, Pearson7Model, StudentsTModel,
                          BreitWignerModel, LognormalModel,
                          DampedOscillatorModel,
                          DampedHarmonicOscillatorModel,
                          ExponentialGaussianModel, SkewedGaussianModel,
                          DonaichModel, PowerLawModel, ExponentialModel,
                          StepModel, RectangleModel, ExpressionModel,
                          update_param_vals)


from .. import Group
from .utils import index_nearest, index_of, savitzky_golay

VALID_BKGS = ('constant', 'linear', 'quadratic')


MODELS = {'constant': ConstantModel,
          'linear': LinearModel,
          'quadratic': QuadraticModel,
          'step': StepModel,
          'rectangle': RectangleModel,
          'exponential': ExponentialModel,
          'gaussian': GaussianModel,
          'lorentzian': LorentzianModel,
          'voigt': VoigtModel,
          'pseudovoigt': PseudoVoigtModel,
          'pearson7': Pearson7Model,
          'dho': DampedHarmonicOscillatorModel,
          'expgaussian': ExponentialGaussianModel,
          'skewedgaussian': SkewedGaussianModel,
          'exponential': ExponentialModel,
          }

# a better guess for step and rectangle models
def step_guess(self, data, x=None, **kwargs):
    if x is None:
        return
    ymin, ymax = min(data), max(data)
    xmin, xmax = min(x), max(x)
    ntest = min(2, len(data)/5)
    step_up =  (data[:ntest].mean() > data[-ntest:].mean())

    dydx = savitzky_golay(np.gradient(data)/np.gradient(x), 5, 2)
    if step_up:
        cen = x[np.where(dydx==dydx.max())][0]
    else:
        cen = x[np.where(dydx==dydx.min())][0]

    pars = self.make_params(amplitude=(ymax-ymin), center=cen)
    pars['%ssigma' % self.prefix].set(value=(xmax-xmin)/5.0, min=0.0)
    return update_param_vals(pars, self.prefix, **kwargs)

def rect_guess(self, data, x=None, **kwargs):
    if x is None:
        return
    ymin, ymax = min(data), max(data)
    xmin, xmax = min(x), max(x)

    ntest = min(2, len(data)/5)
    step_up =  (data[:ntest].mean() > data[-ntest:].mean())


    dydx = savitzky_golay(np.gradient(data)/np.gradient(x), 5, 2)
    cen1 = x[np.where(dydx==dydx.max())][0]
    cen2 = x[np.where(dydx==dydx.min())][0]
    if step_up:
        center1 = cen1 # + (xmax+xmin)/4.0)/2.
        center2 = cen2 # + 3*(xmax+xmin)/4.0)/2.
    else:
        center1 = cen2 # + (xmax+xmin)/4.0)/2.0
        center2 = cen1 # + 3*(xmax+xmin)/4.0)/2.0

    pars = self.make_params(amplitude=(ymax-ymin),
                            center1=center1, center2=center2)

    pars['%ssigma1' % self.prefix].set(value=(xmax-xmin)/5.0, min=0.0)
    pars['%ssigma2' % self.prefix].set(value=(xmax-xmin)/5.0, min=0.0)
    return update_param_vals(pars, self.prefix, **kwargs)

StepModel.guess = step_guess
RectangleModel.guess = rect_guess

def fit_peak(x, y, model, dy=None, background=None, form=None, step=None,
             negative=False, use_gamma=False):
    """fit peak to one a selection of simple 1d models

    out = fit_peak(x, y, model, dy=None,
                   background='linear', form='linear')

    arguments:
    ---------
    x           array of values at which to calculate model
    y           array of values for model to try to match
    dy          array of values for uncertainty in y data to be matched.
    model       name of model to use.  One of (case insensitive)
                     'linear', 'quadratic', 'step', 'rectangle',
                      'gaussian', 'lorentzian', 'voigt', 'exponential'
    background  name of background model to use. One of (case insensitive)
                     None, 'constant', 'linear', or 'quadratic'
                this is ignored when model is 'linear' or 'quadratic'
    form        name of form to use for 'step' and 'rectangle' models.
                One of (case insensitive):
                    'linear', 'erf', or 'atan'
    negative    True/False for whether peak or steps are expected to go down.
    use_gamma   True/False for whether to use separate gamma parameter for
                voigt model.
    output:
    -------
    Group with fit parameters, and more...
    """
    if form is None and step is not None:
        form = step
    out = Group(name='fit_peak result', x=x*1.0, y=y*1.0, dy=1.0,
                model=model, background=background, form=form)

    weight = None
    if dy is not None:
        out.dy = 1.0*dy
        weight = 1.0/max(1.e-16, abs(dy))

    if model.lower() not in MODELS:
        raise ValueError('Unknown fit model: %s ' % model)

    kwargs = dict(negative=negative, background=background,
                  form=form, weight=weight)

    fitclass = MODELS[model.lower()]
    if fitclass == VoigtModel:
        kwargs['use_gamma'] = use_gamma

    mod = fitclass(**kwargs)
    pars = mod.guess(out.y, out.x)

    if background is not None:
        bkg = MODELS[background.lower()](prefix='bkg_')
        bpars = bkg.guess(out.y, x=out.x)
        for p, par  in bpars.items():
            par.value = 0.
            par.vary = True
        pars += bpars
        mod += bkg

    out.init_params = pars

    result = mod.fit(out.y, params=pars, x=out.x) # , dy=out.dy)
    out.fit = mod.eval(result.params, x=out.x)
    out.fit_init = mod.eval(pars, x=out.x)

    out.fit_details = result
    out.chi_square  = result.chisqr
    out.chi_reduced = result.redchi

    for attr in ('aic', 'bic', 'covar', 'rfactor', 'params', 'nvarys',
                 'nfree', 'ndata', 'var_names', 'nfev', 'success',
                 'errorbars', 'message', 'lmdif_message', 'residual'):
        setattr(out, attr, getattr(result, attr, None))

    if background is not None:
        comps = mod.eval_components(x=out.x)
        out.bkg = comps['bkg_']
    return out



#!/usr/bin/env python

import os
import sys
import copy
import os
import time
import gc

import numpy as np
from .. import Group
from ..utils import OrderedDict
from ..utils.strutils import bytes2str

from . import XDIFile, XDIFileException

from .xsp3_hdf5 import XSPRESS3_TAUS, estimate_icr

def read_gsexdi(fname, _larch=None, nmca=128, bad=None, **kws):
    """Read GSE XDI Scan Data to larch group,
    summing ROI data for MCAs and apply deadtime corrections
    """

    MAX_MCAS = nmca
    group = _larch.symtable.create_group()
    group.__name__ ='GSE XDI Data file %s' % fname
    xdi = XDIFile(str(fname))

    group._xdi = xdi
    group.path = fname
    path, suffix = os.path.split(fname)
    group.filename = suffix
    group.npts = xdi.npts
    group.bad_channels = bad

    for family in ('scan', 'mono', 'facility'):
        for key, val in xdi.attrs.get(family, {}).items():
            if '||' in val:
                val, addr = val.split('||')
            try:
                val = float(val)
            except:
                pass
            setattr(group, "%s_%s" % (family, key), val)

    ocrs, icrs = [], []
    ctime = None
    for attrname in dir(xdi):
        if attrname.lower() == 'counttime':
            ctime = getattr(xdi, attrname)
    if ctime is None:
        try:
            ctime = xdi.TSCALER / 5.e7
        except AttributeError:
            ctime = 1.0

    is_xspress3 = any(['13QX4' in a[1] for a in xdi.attrs['column'].items()])
    group.with_xspress3 = is_xspress3
    dtc_taus = XSPRESS3_TAUS
    if _larch.symtable.has_symbol('_sys.gsecars.xspress3_taus'):
        dtc_taus = _larch.symtable._sys.gsecars.xspress3_taus

    dtc_mode = 'icr/ocr'
    for i in range(MAX_MCAS):
        mca = "mca%i" % (i+1)
        ocr    = getattr(xdi, 'OutputCounts_%s' % mca, None)
        clock  = getattr(xdi, 'Clock_%s'        % mca, None)
        icr    = getattr(xdi, 'InputCounts_%s'  % mca, None)
        dtfact = getattr(xdi, 'DTFactor_%s'     % mca, None)
        resets = getattr(xdi, 'ResetTicks_%s'   % mca, None)
        allevt = getattr(xdi, 'AllEvent_%s'     % mca, None)
        if (ocr is None and icr is None and clock is None and
            dtfact is None and resets is None):
            nmca = i
            break
        if ocr is None:
            ocr = ctime
        ocr = ocr/ctime
        if icr is not None:
            icr = icr/ctime

        # Get ICR from one of several alternatives:
        # 1. InputCounts_mca was given
        # 2. DTFactor_mca was given
        # 3. ResetTicks_mca and AllEvets_mca were given
        # 4. Use "known values" for tau
        if icr is None:
            if dtfact is not None:
                icr = ocr * dtfact
                dtc_mode = 'dtfactor'
            elif (clock is not None and
                  resets is not None and
                  allevt is not None):
                dtfact = clock/(clock - (6*allevt + resets))
                icr = ocr * dtfact
                dtc_mode = 'resets'
        # finally estimate from measured values of tau:
        if icr is None:
            icr = 1.0*ocr
            dtc_mode = 'none'
            if is_xspress3:
                tau = dtc_taus[i]
                icr = estimate_icr(ocr*1.00, tau, niter=7)
                dtc_mode = 'saved_taus'
        ocrs.append(ocr)
        icrs.append(icr)

    group.dtc_mode =  dtc_mode
    if dtc_mode == 'saved_taus':
        group.dtc_taus = dtc_taus

    labels = []
    sums = OrderedDict()
    for i, arrname in enumerate(xdi.array_labels):
        dat = getattr(xdi, arrname)
        if arrname.lower() == 'data':
            arrname = '_data'
        aname = sumname = rawname = arrname.lower()
        if ('_mca' in aname and
            'outputcounts' not in aname and
            'dtfactor' not in aname and
            'clock' not in aname):
            sumname, imca = sumname.split('_mca')
            imca = int(imca) - 1
            datraw = dat*1.0
            rawname = sumname + '_nodtc'
            dat   = dat * icrs[imca]/ ocrs[imca]
            if any(np.isnan(dat)):
                nan_pts = np.where(np.isnan(dat))[0]
                dat[nan_pts] = datraw[nan_pts]

        setattr(group, aname, dat)
        if sumname not in labels:
            labels.append(sumname)
            sums[sumname] = dat
            if rawname != sumname:
                sums[rawname] = datraw
                if rawname not in labels:
                    labels.append(rawname)

        else:
            sums[sumname] = sums[sumname] + dat
            if rawname != sumname:
                sums[rawname] = sums[rawname] + datraw

    for name, dat in sums.items():
        if not hasattr(group, name):
            setattr(group, name, dat)

    for arrname in xdi.array_labels:
        sname = arrname.lower()
        if sname not in labels:
            labels.append(sname)

    data = []
    for name in labels:
        data.append(getattr(group, name))
    group.data = np.array(data)
    for imca in range(nmca):
        setattr(group, 'ocr_mca%i' % (imca+1), ocrs[imca])
        setattr(group, 'icr_mca%i' % (imca+1), icrs[imca])
    group.array_labels = labels
    return group


DTC_header = '''# XDI/1.1  Epics StepScan File/2.0
# Beamline.name:  13-ID-E, GSECARS
# Monochromator.name:  %(mono_cut)s, LN2 Cooled
# Monochromator.dspacing:  %(mono_dspace)s
# Facility.name: APS
# Facility.xray_source: 3.6 cm undulator
# Detectors.i0:  20cm ion chamber, He
# Detectors.ifluor:  Si SDD Vortex ME-4, 4 elements
# Detectors.ifluor_electronics:  Quantum Xspress3 3.1.10'''

def is_GSEXDI(filename):
    """test if file is GSE XDI data file
    reads only the first line of file
    """
    line1 = open(filename, 'r').readline()
    return (line1.startswith('#XDI/1') and 'Epics StepScan File' in line1)

def gsexdi_deadtime_correct(fname, channelname, subdir='DT_Corrected',
                            bad=None, _larch=None):
    """convert GSE XDI fluorescence XAFS scans to dead time corrected files"""
    if not is_GSEXDI(fname):
        print("'%s' is not a GSE XDI scan file\n" % fname)
        return

    out = Group()
    out.orig_filename = fname
    try:
        xdi = read_gsexdi(fname, bad=bad, _larch=_larch)
    except:
        print('Could not read XDI file ', fname)
        return

    for attr in ('energy', 'i0', 'i1', 'i2', 'tscaler',
                 'counttime',  'scan_start_time', 'scan_end_time'):
        if hasattr(xdi, attr):
            setattr(out, attr, getattr(xdi, attr))

    # some scans may not record separate counttime, but TSCALER
    # is clock ticks for a 50MHz clock
    if not hasattr(out, 'counttime'):
        out.counttime = xdi.tscaler * 2.e-8

    if hasattr(xdi, 'energy_readback'):
        out.energy = xdi.energy_readback

    arrname = None
    channelname = channelname.lower().replace(' ', '_')

    for arr in xdi.array_labels:
        if arr.lower().startswith(channelname):
            arrname = arr
            break
    if arrname is None:
        print('Cannot find Channel %s in file %s '% (channelname, fname))
        return

    out.ifluor = getattr(xdi, arrname)
    out.ifluor_raw  = getattr(xdi, arrname)
    arrname_raw = arrname + '_nodtc'
    if arrname_raw  in xdi.array_labels:
        out.ifluor_raw  = getattr(xdi, arrname_raw)

    out.mufluor = out.ifluor / out.i0
    TINY  = 2.e-20
    if hasattr(out, 'i1') or hasattr(out, 'itrans'):
        i1 = getattr(out, 'i1', None)
        if i1 is None:
            i1 = getattr(out, 'itrans', None)
        if i1 is not None:
            i1[np.isnan(i1)] = TINY
            i1 = i1 / out.i0
            i1[np.where(i1<TINY)] = TINY
            out.mutrans = -np.log(i1)


    npts   = len(out.i0)
    col0_name = xdi.array_labels[0].lower()
    col0_units = None
    col0_data  = xdi.data[0, :]
    if col0_name == 'energy':
        col0_data  = out.energy
        col0_units = 'eV'

    buff =  ['# XDI/1.0  GSE/1.0']

    header = OrderedDict()

    hgroups = ['beamline', 'facility', 'mono', 'undulator', 'detectors',
               'scaler', 'detectorstage', 'samplestage', 'scan', 'scanparameters']
    hskip = ['scanparameters.end', 'scanparameters.start']
    for agroup in hgroups:
        attrs = xdi._xdi.attrs.get(agroup, {})
        if agroup == 'mono': agroup = 'monochromator'
        header[agroup] = OrderedDict()
        for sname in sorted(attrs.keys()):
            if "%s.%s" %( agroup, sname) not in hskip:
                header[agroup][sname] = attrs[sname]


    header['facility']['name'] = 'APS'
    header['facility']['xray_source'] = '3.6 cm undulator'
    header['beamline']['name'] = '13-ID-E, GSECARS'

    header['detectors']['i0'] = '20cm ion chamber, He'
    header['detectors']['ifluor'] = 'Si SDD Vortex ME-4, 4 elements'
    header['detectors']['ifluor_electronics'] = 'Quantum Xspress3 3.1.10'

    mono_cut = 'Si(111)'
    if xdi.mono_dspacing < 2:
        mono_cut = 'Si(311)'
    header['monochromator']['name'] = "%s, LN2 cooled"  % mono_cut

    out_arrays = OrderedDict()
    out_arrays[col0_name]  = (col0_name, col0_units)
    out_arrays['mufluor'] = ('mufluor', None)
    if hasattr(out, 'i1'):
        out_arrays['mutrans'] = ('mutrans', None)

    out_arrays['ifluor']  = ('ifluor', '# deadtime-corrected')
    out_arrays['ifluor_raw'] = ('ifluor_raw', '# not deadtime-corrected')
    out_arrays['i0'] = ('i0', None)

    if hasattr(out, 'i1'):
        out_arrays['itrans'] = ('i1', None)
    if hasattr(out, 'i2'):
        out_arrays['irefer'] = ('i2', None)

    if hasattr(out, 'counttime'):
        out_arrays['counttime'] = ('counttime', 'sec')

    arrlabel = []
    for iarr, aname in enumerate(out_arrays):
        lab = "%12s " % aname
        if iarr == 0: lab = "%11s " % aname
        arrlabel.append(lab)
        extra = out_arrays[aname][1]
        if extra is None: extra = ''
        buff.append("# Column.%i: %s %s" % (iarr+1, aname, extra))

    arrlabel = '#%s' % (' '.join(arrlabel))
    ncol = len(out_arrays)

    for family, fval in header.items():
        for attr, val in fval.items():
            buff.append("# %s.%s: %s" % (family.title(), attr, val))


    buff.append("# ///")
    for comment in bytes2str(xdi._xdi.comments).split('\n'):
        c = comment.strip()
        if len(c) > 0:
            buff.append('# %s' % c)
    buff.extend(["# summed %s fluorescence data from %s" % (channelname, fname),
                 "# Dead-time correction applied",
                 "#"+ "-"*78,   arrlabel])

    efmt = "%11.4f"
    ffmt = "%13.7f"
    gfmt = "%13.7g"
    for i in range(npts):
        dline = ["", efmt % col0_data[i], ffmt % out.mufluor[i]]

        if hasattr(out, 'i1'):
            dline.append(ffmt % out.mutrans[i])

        dline.extend([gfmt % out.ifluor[i],
                      gfmt % out.ifluor_raw[i],
                      gfmt % out.i0[i]])

        if hasattr(out, 'i1'):
            dline.append(gfmt % out.i1[i])
        if hasattr(out, 'i2'):
            dline.append(gfmt % out.i2[i])
        if hasattr(out, 'counttime'):
            dline.append(gfmt % out.counttime[i])

        buff.append(" ".join(dline))
    ofile = fname[:]
    if ofile.startswith('..'):
        ofile = ofile[3:]
    ofile = ofile.replace('.', '_') + '.dat'
    ofile = os.path.join(subdir, ofile)
    if not os.path.exists(subdir):
        os.mkdir(subdir)
    try:
       fout = open(ofile, 'w')
       fout.write("\n".join(buff))
       fout.close()
       print("wrote %s, npts=%i, channel='%s'" % (ofile, npts, channelname))
    except:
       print("could not open / write to output file %s" % ofile)

    return out
